import os
from contextlib import contextmanager
from datetime import date, timedelta
from tempfile import TemporaryDirectory
from unittest import mock

import pytest
import responses
from sqlalchemy import create_engine

from cumulusci.core.exceptions import (
    BulkDataException,
    ConfigError,
    CumulusCIException,
    TaskOptionsError,
)
from cumulusci.tasks.bulkdata import ExtractData
from cumulusci.tasks.bulkdata.mapping_parser import MappingLookup, MappingStep
from cumulusci.tasks.bulkdata.step import (
    BaseQueryOperation,
    DataApi,
    DataOperationJobResult,
    DataOperationStatus,
    DataOperationType,
)
from cumulusci.tasks.bulkdata.tests.utils import _make_task
from cumulusci.tests.util import (
    assert_max_memory_usage,
    mock_describe_calls,
    mock_salesforce_client,
)
from cumulusci.utils import temporary_dir


@contextmanager
def mock_extract_jobs(task, extracted_records):
    def _job_state_from_batches(self, job_id):
        return DataOperationJobResult(
            DataOperationStatus.SUCCESS,
            [],
            10,
            0,
        )

    def get_results(self):
        return extracted_records[self.sobject]

    with mock.patch(
        "cumulusci.tasks.bulkdata.step.BulkApiQueryOperation.get_results",
        get_results,
    ), mock.patch(
        "cumulusci.tasks.bulkdata.step.BulkJobMixin._job_state_from_batches",
        _job_state_from_batches,
    ):
        yield


class MockBulkQueryOperation(BaseQueryOperation):
    def __init__(self, *, sobject, api_options, context, query):
        super().__init__(
            sobject=sobject, api_options=api_options, context=context, query=query
        )
        self.results = []

    def query(self):
        self.job_id = "JOB"
        self.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], len(self.results), 0
        )

    def get_results(self):
        return iter(self.results)


class MockScalableBulkQueryOperation(MockBulkQueryOperation):
    def query(self):
        self.job_id = "JOB"
        self.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], self.result_len, 0
        )


class TestExtractData:

    mapping_file_v1 = "mapping_v1.yml"
    mapping_file_v2 = "mapping_v2.yml"
    mapping_file_poly = "mapping_poly.yml"
    mapping_file_poly_wrong = "mapping_poly_wrong.yml"
    mapping_file_poly_incomplete = "mapping_poly_incomplete.yml"
    mapping_file_vanilla = "mapping_vanilla_sf.yml"

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run__person_accounts_disabled(self, query_op_mock):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        mock_describe_calls()
        with temporary_dir() as d:
            tmp_db_path = os.path.join(d, "testdata.db")

            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{tmp_db_path}",
                        "mapping": mapping_path,
                    }
                },
            )
            task.bulk = mock.Mock()
            task.sf = mock.Mock()
            task.org_config._is_person_accounts_enabled = False

            mock_query_households = MockBulkQueryOperation(
                sobject="Account",
                api_options={},
                context=task,
                query="SELECT Id FROM Account",
            )
            mock_query_contacts = MockBulkQueryOperation(
                sobject="Contact",
                api_options={},
                context=task,
                query="SELECT Id, FirstName, LastName, Email, AccountId FROM Contact",
            )
            mock_query_households.results = [["1"]]
            mock_query_contacts.results = [
                ["2", "First", "Last", "test@example.com", "1"]
            ]

            query_op_mock.side_effect = [mock_query_households, mock_query_contacts]

            task()

            with create_engine(task.options["database_url"]).connect() as conn:
                household = next(conn.execute("select * from households"))
                assert household.sf_id == "1"
                assert not hasattr(household, "IsPersonAccount")
                assert household.record_type == "HH_Account"

                contact = next(conn.execute("select * from contacts"))
                assert contact.sf_id == "2"
                assert not hasattr(contact, "IsPersonAccount")
                assert contact.household_id == "1"

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run__person_accounts_enabled(self, query_op_mock):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        mock_describe_calls()

        with temporary_dir() as d:
            tmp_db_path = os.path.join(d, "testdata.db")

            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{tmp_db_path}",
                        "mapping": mapping_path,
                    }
                },
            )
            task.bulk = mock.Mock()
            task.sf = mock.Mock()
            task.org_config._is_person_accounts_enabled = True

            mock_query_households = MockBulkQueryOperation(
                sobject="Account",
                api_options={},
                context=task,
                query="SELECT Id, IsPersonAccount FROM Account",
            )
            mock_query_contacts = MockBulkQueryOperation(
                sobject="Contact",
                api_options={},
                context=task,
                query="SELECT Id, FirstName, LastName, Email, IsPersonAccount, AccountId FROM Contact",
            )
            mock_query_households.results = [["1", "false"]]
            mock_query_contacts.results = [
                ["2", "First", "Last", "test@example.com", "true", "1"]
            ]

            query_op_mock.side_effect = [mock_query_households, mock_query_contacts]

            task()
            with create_engine(task.options["database_url"]).connect() as conn:

                household = next(conn.execute("select * from households"))
                assert household.sf_id == "1"
                assert household.IsPersonAccount == "false"
                assert household.record_type == "HH_Account"

                contact = next(conn.execute("select * from contacts"))
                assert contact.sf_id == "2"
                assert contact.IsPersonAccount == "true"
                assert contact.household_id == "1"

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run__sql(self, query_op_mock):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        mock_describe_calls()

        with temporary_dir():
            task = _make_task(
                ExtractData,
                {"options": {"sql_path": "testdata.sql", "mapping": mapping_path}},
            )
            task.bulk = mock.Mock()
            task.sf = mock.Mock()
            task.org_config._is_person_accounts_enabled = False

            mock_query_households = MockBulkQueryOperation(
                sobject="Account",
                api_options={},
                context=task,
                query="SELECT Id FROM Account",
            )
            mock_query_contacts = MockBulkQueryOperation(
                sobject="Contact",
                api_options={},
                context=task,
                query="SELECT Id, FirstName, LastName, Email, AccountId FROM Contact",
            )
            mock_query_households.results = [["1"]]
            mock_query_contacts.results = [
                ["2", "First☃", "Last", "test@example.com", "1"]
            ]
            query_op_mock.side_effect = [mock_query_households, mock_query_contacts]

            with mock.patch(
                "cumulusci.tasks.bulkdata.extract.create_engine", wraps=create_engine
            ) as ce_mock:
                task()

            assert os.path.exists("testdata.sql")
            assert ce_mock.mock_calls[0][1][0].endswith(
                "temp_db.db"
            ), ce_mock.mock_calls[0][1][0]
            assert ce_mock.mock_calls[0][1][0].startswith(
                "sqlite:///"
            ), ce_mock.mock_calls[0][1][0]

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run__v2__person_accounts_disabled(self, query_op_mock):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v2)
        mock_describe_calls()

        with TemporaryDirectory() as t:
            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{t}/temp_db",  # in memory
                        "mapping": mapping_path,
                    }
                },
            )
            task.bulk = mock.Mock()
            task.sf = mock.Mock()
            task.org_config._is_person_accounts_enabled = False

            mock_query_households = MockBulkQueryOperation(
                sobject="Account",
                api_options={},
                context=task,
                query="SELECT Id, Name FROM Account",
            )
            mock_query_contacts = MockBulkQueryOperation(
                sobject="Contact",
                api_options={},
                context=task,
                query="SELECT Id, FirstName, LastName, Email, AccountId FROM Contact",
            )
            mock_query_households.results = [["1", "TestHousehold"]]
            mock_query_contacts.results = [
                ["2", "First", "Last", "test@example.com", "1"]
            ]

            query_op_mock.side_effect = [mock_query_households, mock_query_contacts]

            task()
            with create_engine(task.options["database_url"]).connect() as conn:
                household = next(conn.execute("select * from households"))
                assert household.name == "TestHousehold"
                assert not hasattr(household, "IsPersonAccount")
                assert household.record_type == "HH_Account"

                contact = next(conn.execute("select * from contacts"))
                assert contact.household_id == "Account-1"
                assert not hasattr(contact, "IsPersonAccount")

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run__v2__person_accounts_enabled(self, query_op_mock):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v2)
        mock_describe_calls()
        with temporary_dir() as d:
            tmp_db_path = os.path.join(d, "testdata.db")
            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{tmp_db_path}",
                        "mapping": mapping_path,
                    }
                },
            )
            task.bulk = mock.Mock()
            task.sf = mock.Mock()
            task.org_config._is_person_accounts_enabled = True

            mock_query_households = MockBulkQueryOperation(
                sobject="Account",
                api_options={},
                context=task,
                query="SELECT Id, Name, IsPersonAccount FROM Account",
            )
            mock_query_contacts = MockBulkQueryOperation(
                sobject="Contact",
                api_options={},
                context=task,
                query="SELECT Id, FirstName, LastName, Email, IsPersonAccount, AccountId FROM Contact",
            )
            mock_query_households.results = [["1", "TestHousehold", "false"]]
            mock_query_contacts.results = [
                ["2", "First", "Last", "test@example.com", "true", "1"]
            ]

            query_op_mock.side_effect = [mock_query_households, mock_query_contacts]

            task()
            with create_engine(task.options["database_url"]).connect() as conn:
                household = next(conn.execute("select * from households"))
                assert household.name == "TestHousehold"
                assert household.IsPersonAccount == "false"
                assert household.record_type == "HH_Account"

                contact = next(conn.execute("select * from contacts"))
                assert contact.household_id == "Account-1"
                assert contact.IsPersonAccount == "true"

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run__poly__polymorphic_lookups(self, query_op_mock):
        """Test for polymorphic lookups"""
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_poly)
        mock_describe_calls()

        with temporary_dir() as t:
            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{t}/temp_poly.db",  # in memory
                        "mapping": mapping_path,
                    }
                },
            )
            task.bulk = mock.Mock()
            task.sf = mock.Mock()
            task.org_config._is_person_accounts_enabled = False

            mock_query_households = MockBulkQueryOperation(
                sobject="Account",
                api_options={},
                context=task,
                query="SELECT Id, Name FROM Account",
            )
            mock_query_contacts = MockBulkQueryOperation(
                sobject="Contact",
                api_options={},
                context=task,
                query="SELECT Id, FirstName, LastName, Email, AccountId FROM Contact",
            )
            mock_query_events = MockBulkQueryOperation(
                sobject="Event",
                api_options={},
                context=task,
                query="SELECT Id, LastName, WhoId FROM Event",
            )
            mock_query_households.results = [["abc123", "TestHousehold"]]
            mock_query_contacts.results = [
                ["def456", "First", "Last", "test@example.com", "abc123"]
            ]
            mock_query_events.results = [
                ["ijk789", "Last1", "abc123"],
                ["lmn010", "Last2", "def456"],
            ]

            query_op_mock.side_effect = [
                mock_query_households,
                mock_query_contacts,
                mock_query_events,
            ]
            task()
            with create_engine(task.options["database_url"]).connect() as conn:
                household = next(conn.execute("select * from households"))
                assert household.name == "TestHousehold"
                assert not hasattr(household, "IsPersonAccount")
                assert household.record_type == "HH_Account"

                contact = next(conn.execute("select * from contacts"))
                assert contact.household_id == "Account-1"
                assert not hasattr(contact, "IsPersonAccount")

                events = conn.execute("select * from events").fetchall()
                assert events[0].who_id == "Account-1"
                assert events[1].who_id == "Contact-1"

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run__poly__wrong_mapping(self, query_op_mock):
        """Test for polymorphic lookups with wrong mapping file
        (missing table)"""
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_poly_wrong)
        mock_describe_calls()

        with temporary_dir() as t:
            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{t}/temp_poly.db",  # in memory
                        "mapping": mapping_path,
                    }
                },
            )
            task.bulk = mock.Mock()
            task.sf = mock.Mock()
            task.org_config._is_person_accounts_enabled = False

            mock_query_households = MockBulkQueryOperation(
                sobject="Account",
                api_options={},
                context=task,
                query="SELECT Id, Name FROM Account",
            )
            mock_query_contacts = MockBulkQueryOperation(
                sobject="Contact",
                api_options={},
                context=task,
                query="SELECT Id, FirstName, LastName, Email, AccountId FROM Contact",
            )
            mock_query_events = MockBulkQueryOperation(
                sobject="Event",
                api_options={},
                context=task,
                query="SELECT Id, LastName, WhoId FROM Event",
            )
            mock_query_households.results = [["abc123", "TestHousehold"]]
            mock_query_contacts.results = [
                ["def456", "First", "Last", "test@example.com", "abc123"]
            ]
            mock_query_events.results = [
                ["ijk789", "Last1", "abc123"],
                ["lmn010", "Last2", "def456"],
            ]

            query_op_mock.side_effect = [
                mock_query_households,
                mock_query_contacts,
                mock_query_events,
            ]
            with pytest.raises(CumulusCIException) as e:
                task()

            assert "The following tables are missing in the mapping file:" in str(
                e.value
            )

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run__poly__incomplete_mapping(self, query_op_mock):
        """Test for polymorphic lookups with incomplete mapping file"""
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_poly_incomplete)
        mock_describe_calls()

        with temporary_dir() as t:
            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{t}/temp_poly.db",  # in memory
                        "mapping": mapping_path,
                    }
                },
            )
            task.bulk = mock.Mock()
            task.sf = mock.Mock()
            task.org_config._is_person_accounts_enabled = False

            mock_query_households = MockBulkQueryOperation(
                sobject="Account",
                api_options={},
                context=task,
                query="SELECT Id, Name FROM Account",
            )
            mock_query_events = MockBulkQueryOperation(
                sobject="Event",
                api_options={},
                context=task,
                query="SELECT Id, LastName, WhoId FROM Event",
            )
            mock_query_households.results = [["abc123", "TestHousehold"]]
            mock_query_events.results = [
                ["ijk789", "Last1", "abc123"],
                ["lmn010", "Last2", "def456"],
            ]

            query_op_mock.side_effect = [mock_query_households, mock_query_events]
            with pytest.raises(ConfigError) as e:
                task()

            assert "Total mapping operations" in str(e.value)
            assert "do not match total non-empty rows" in str(e.value)

    @mock.patch("cumulusci.tasks.bulkdata.extract.log_progress")
    def test_import_results__oid_as_pk(self, log_mock):
        task = _make_task(
            ExtractData,
            {"options": {"database_url": "sqlite://", "mapping": "mapping.yml"}},
        )

        mapping = MappingStep(
            sf_object="Opportunity",
            fields={"Id": "sf_id", "Name": "Name"},
            lookups={"AccountId": MappingLookup(table="Account", name="AccountId")},
        )
        step = mock.Mock()
        task.session = mock.Mock()
        task.metadata = mock.MagicMock()

        with mock.patch(
            "cumulusci.tasks.bulkdata.extract.sql_bulk_insert_from_records"
        ) as sql_bulk_insert_from_records:
            task._import_results(mapping, step)

        task.session.connection.assert_called_once_with()
        step.get_results.assert_called_once_with()
        sql_bulk_insert_from_records.assert_called_once_with(
            connection=task.session.connection.return_value,
            table=task.metadata.tables[mapping.table],
            columns=["sf_id", "Name", "AccountId"],
            record_iterable=log_mock.return_value,
        )

    @responses.activate
    def test_import_results__no_columns(self):  # , query_op_mock):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        mock_describe_calls()

        task = _make_task(
            ExtractData,
            {"options": {"database_url": "sqlite://", "mapping": mapping_path}},
        )

        mapping = MappingStep(
            sf_object="Opportunity",
            table="Opportunity",
            fields={},
            lookups={},
        )
        step = mock.Mock()
        step.get_results.return_value = [[1], [2]]
        task.session = mock.Mock()
        task._init_task()
        task._init_mapping()
        task.mapping["Opportunity"] = mapping
        with task._init_db():
            task._import_results(mapping, step)
            output_Opportunties = list(
                task.session.execute("select * from Opportunity")
            )
            assert output_Opportunties == [("Opportunity-1",), ("Opportunity-2",)]

    @responses.activate
    def test_import_results__relative_dates(self):
        mock_describe_calls()
        task = _make_task(
            ExtractData,
            {"options": {"database_url": "sqlite://", "mapping": "mapping.yml"}},
        )

        mapping = MappingStep(
            sf_object="Opportunity",
            fields={"Id": "sf_id", "CloseDate": "CloseDate"},
            anchor_date="2020-07-01",
        )
        step = mock.Mock()
        step.get_results.return_value = iter(
            [["006000000000001", (date.today() + timedelta(days=9)).isoformat()]]
        )
        task.session = mock.Mock()
        task.metadata = mock.MagicMock()
        task._init_task()
        with mock.patch(
            "cumulusci.tasks.bulkdata.extract.sql_bulk_insert_from_records"
        ) as sql_bulk_insert_from_records:
            task._import_results(mapping, step)

        task.session.connection.assert_called_once_with()
        step.get_results.assert_called_once_with()
        sql_bulk_insert_from_records.assert_called_once_with(
            connection=task.session.connection.return_value,
            table=task.metadata.tables[mapping.table],
            columns=["sf_id", "CloseDate"],
            record_iterable=mock.ANY,
        )

        result = list(
            sql_bulk_insert_from_records.call_args_list[0][1]["record_iterable"]
        )
        assert result == [["006000000000001", "2020-07-10"]]

    def test_import_results__record_type_mapping(self):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, "recordtypes.yml")
        task = _make_task(
            ExtractData,
            {"options": {"database_url": "sqlite://", "mapping": mapping_path}},
        )
        task._extract_record_types = mock.Mock()
        task.metadata = mock.MagicMock()
        task.session = mock.Mock()
        task.org_config._is_person_accounts_enabled = False

        step = mock.Mock()
        step.get_results.return_value = [["000000000000001", "Test", "012000000000000"]]

        mapping = MappingStep(
            sf_object="Account",
            fields={"Name": "Name", "RecordTypeId": "RecordTypeId"},
            lookups={},
            table="accounts",
        )
        with mock.patch(
            "cumulusci.tasks.bulkdata.extract.sql_bulk_insert_from_records_incremental",
            return_value=[None, None],
        ):
            task._import_results(
                mapping,
                step,
            )
        task._extract_record_types.assert_called_once_with(
            "Account",
            mapping.get_source_record_type_table(),
            task.session.connection.return_value,
            task.org_config._is_person_accounts_enabled,
        )

    def test_import_results__person_account_name_stripped(self):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, "recordtypes.yml")
        task = _make_task(
            ExtractData,
            {"options": {"database_url": "sqlite://", "mapping": mapping_path}},
        )
        task._extract_record_types = mock.Mock()
        task.session = mock.Mock()
        task.metadata = mock.MagicMock()
        task.org_config._is_person_accounts_enabled = True

        step = mock.Mock()
        step.get_results.return_value = [
            ["000000000000001", "Person Account", "012000000000001", "true"],
            ["000000000000002", "Business Account", "012000000000002", "false"],
        ]

        with mock.patch(
            "cumulusci.tasks.bulkdata.extract.sql_bulk_insert_from_records"
        ) as sql_bulk_insert_from_records:
            task._import_results(
                MappingStep(
                    sf_object="Account",
                    fields={
                        "Id": "sf_id",
                        "Name": "Name",
                        "RecordTypeId": "RecordTypeId",
                        "IsPersonAccount": "IsPersonAccount",
                    },
                    lookups={},
                ),
                step,
            )

        sql_bulk_insert_from_records.assert_called()
        args, kwargs = sql_bulk_insert_from_records.call_args_list[0]

        records = [record for record in kwargs["record_iterable"]]

        assert [
            ["000000000000001", "", "012000000000001", "true"],
            ["000000000000002", "Business Account", "012000000000002", "false"],
        ] == records

    @responses.activate
    def test_map_autopks(self):
        mock_describe_calls()
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v2)
        mock_describe_calls()

        task = _make_task(
            ExtractData,
            {
                "options": {
                    "database_url": "sqlite://",  # in memory
                    "mapping": mapping_path,
                }
            },
        )
        task._convert_lookups_to_id = mock.Mock()
        task.metadata = mock.MagicMock()

        task._init_task()
        task._init_mapping()
        task._map_autopks()

        task._convert_lookups_to_id.assert_called_once_with(
            task.mapping["Insert Contacts"], ["AccountId"]
        )
        task.metadata.tables.__getitem__.return_value.drop.assert_has_calls(
            [mock.call(), mock.call()]
        )
        task.metadata.tables.__getitem__.assert_has_calls(
            [mock.call("contacts_sf_ids"), mock.call("households_sf_ids")],
            any_order=True,
        )

    def test_convert_lookups_to_id(self):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )

        task.session = mock.Mock()
        task.models = {
            "Account": mock.Mock(),
            "Account_sf_ids": mock.Mock(),
            "Opportunity": mock.Mock(),
            "Opportunity_sf_ids": mock.Mock(),
        }
        task.mapping = {
            "Account": MappingStep(sf_object="Account"),
            "Opportunity": MappingStep(sf_object="Opportunity"),
        }

        task.session.query.return_value.filter.return_value.count.return_value = 0
        task.session.query.return_value.filter.return_value.update.return_value.rowcount = (
            0
        )
        task._convert_lookups_to_id(
            MappingStep(
                sf_object="Opportunity",
                lookups={"AccountId": MappingLookup(table="Account", name="AccountId")},
            ),
            ["AccountId"],
        )

        task.session.query.return_value.filter.return_value.update.assert_called_once_with(
            {task.models["Opportunity"].AccountId: task.models["Account_sf_ids"].id},
            synchronize_session=False,
        )
        task.session.commit.assert_called_once_with()

    def test_convert_lookups_to_id__sqlite(self):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )

        task.session = mock.Mock()
        task.models = {
            "Account": mock.Mock(),
            "Account_sf_ids": mock.Mock(),
            "Opportunity": mock.Mock(),
            "Opportunity_sf_ids": mock.Mock(),
        }
        task.mapping = {
            "Account": MappingStep(sf_object="Account"),
            "Opportunity": MappingStep(sf_object="Opportunity"),
        }
        task.session.query.return_value.filter.return_value.update.side_effect = (
            NotImplementedError
        )

        item = mock.Mock()

        task.session.query.return_value.join.return_value = [(item, "1")]
        task.session.query.return_value.filter.return_value.count.return_value = 1

        task._convert_lookups_to_id(
            MappingStep(
                sf_object="Opportunity",
                lookups={"AccountId": MappingLookup(table="Account", name="AccountId")},
            ),
            ["AccountId"],
        )

        task.session.bulk_update_mappings.assert_called_once_with(
            task.models["Opportunity"], [{"id": item.id, "AccountId": "1"}]
        )
        task.session.commit.assert_called_once_with()

    @mock.patch("cumulusci.tasks.bulkdata.extract.create_table")
    @mock.patch("cumulusci.tasks.bulkdata.extract.mapper")
    def test_create_table(self, mapper_mock, create_mock):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )
        mapping = MappingStep(
            sf_object="Account",
            fields={"Name": "Name", "Id": "sf_id"},
            lookups={},
            table="accounts",
        )
        task.models = {}
        task.metadata = mock.Mock()
        task._create_table(mapping)
        create_mock.assert_called_once_with(mapping, task.metadata)

        assert "accounts" in task.models

    @responses.activate
    def test_create_table__already_exists(self):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        db_path = os.path.join(base_path, "testdata.db")
        mock_describe_calls()
        task = _make_task(
            ExtractData,
            {
                "options": {
                    "database_url": f"sqlite:///{db_path}",
                    "mapping": mapping_path,
                }
            },
        )
        task.org_config._is_person_accounts_enabled = False

        with pytest.raises(BulkDataException):
            task()

    def test_create_table__record_type_mapping(self):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )
        task.mapping = {
            "Insert Accounts": MappingStep(
                sf_object="Account",
                table="accounts",
                fields={"Name": "Name", "RecordTypeId": "RecordTypeId"},
                lookups={},
            ),
            "Insert Other Accounts": MappingStep(
                sf_object="Account",
                fields={"Name": "Name", "RecordTypeId": "RecordTypeId"},
                lookups={},
                table="accounts_2",
            ),
        }
        task.org_config._is_person_accounts_enabled = False

        def create_table_mock(table_name):
            task.models[table_name] = mock.Mock()

        task._create_record_type_table = mock.Mock(side_effect=create_table_mock)
        with task._init_db():
            task._create_record_type_table.assert_called_once_with("Account_rt_mapping")

    @mock.patch("cumulusci.tasks.bulkdata.extract.create_table")
    @mock.patch("cumulusci.tasks.bulkdata.extract.Table")
    @mock.patch("cumulusci.tasks.bulkdata.extract.mapper")
    def test_create_table__autopk(self, mapper_mock, table_mock, create_mock):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )
        mapping = MappingStep(
            sf_object="Account",
            fields={"Name": "Name"},
            table="accounts",
        )
        task.models = {}
        task.metadata = mock.Mock()
        task.org_config._is_person_accounts_enabled = False

        task._create_table(mapping)

        create_mock.assert_called_once_with(mapping, task.metadata)
        assert len(table_mock.mock_calls) == 1

        assert "accounts" in task.models
        assert mapping.get_sf_id_table() in task.models

    def test_create_tables(self):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )
        task.mapping = {1: "test", 2: "foo", 3: "bar"}
        task.metadata = mock.Mock()
        task._create_table = mock.Mock()

        task._create_tables()

        task._create_table.assert_has_calls(
            [mock.call("test"), mock.call("foo"), mock.call("bar")]
        )
        task.metadata.create_all.assert_called_once_with()

    def test_init_db(self):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )
        task._create_tables = mock.Mock()
        with task._init_db():
            assert task.models == {}
            assert task.session.query

    def assert_person_accounts_in_mapping(
        self, mapping, org_has_person_accounts_enabled
    ):
        for step in mapping.values():
            if step["sf_object"] in ("Account", "Contact"):
                assert org_has_person_accounts_enabled == (
                    "IsPersonAccount" in step["fields"]
                )

    @responses.activate
    def test_init_mapping(self):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        mock_describe_calls()
        task = _make_task(
            ExtractData,
            {"options": {"database_url": "sqlite:///", "mapping": mapping_path}},
        )
        task.org_config._is_person_accounts_enabled = False

        task._init_task()
        task._init_mapping()
        assert "Insert Households" in task.mapping

        # Person Accounts should not be added to mapping
        self.assert_person_accounts_in_mapping(task.mapping, False)

    @responses.activate
    def test_init_mapping_org_has_person_accounts_enabled(self):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        mock_describe_calls()
        task = _make_task(
            ExtractData,
            {"options": {"database_url": "sqlite:///", "mapping": mapping_path}},
        )
        task.org_config._is_person_accounts_enabled = True

        task._init_task()
        task._init_mapping()
        assert "Insert Households" in task.mapping

        # Person Accounts should not be added to mapping
        self.assert_person_accounts_in_mapping(task.mapping, True)

    @mock.patch("cumulusci.tasks.bulkdata.extract.validate_and_inject_mapping")
    def test_init_mapping_passes_options_to_validate(self, validate_and_inject_mapping):
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        t = _make_task(
            ExtractData,
            {
                "options": {
                    "database_url": "sqlite:///",
                    "mapping": mapping_path,
                    "inject_namespaces": True,
                    "drop_missing_schema": True,
                }
            },
        )
        t.org_config._is_person_accounts_enabled = True

        t._init_task()
        t._init_mapping()

        validate_and_inject_mapping.assert_called_once_with(
            mapping=t.mapping,
            sf=t.sf,
            namespace=t.project_config.project__package__namespace,
            data_operation=DataOperationType.QUERY,
            inject_namespaces=True,
            drop_missing=True,
            org_has_person_accounts_enabled=t.org_config._is_person_accounts_enabled,
        )

    def test_soql_for_mapping(self):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )
        mapping = MappingStep(
            sf_object="Contact",
            fields={"Id": "sf_id", "Test__c": "Test"},
        )
        assert task._soql_for_mapping(mapping) == "SELECT Id, Test__c FROM Contact"

        mapping = MappingStep(
            sf_object="Contact",
            record_type="Devel",
            fields={"Id": "sf_id", "Test__c": "Test"},
        )
        assert (
            task._soql_for_mapping(mapping)
            == "SELECT Id, Test__c FROM Contact WHERE RecordType.DeveloperName = 'Devel'"
        )

    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run_query(self, query_op_mock):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )
        task._import_results = mock.Mock()
        query_op_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 1, 0
        )

        task._run_query("SELECT Id FROM Contact", MappingStep(sf_object="Contact"))

        query_op_mock.assert_called_once_with(
            sobject="Contact",
            fields=["Id"],
            api=DataApi.SMART,
            api_options={},
            context=task,
            query="SELECT Id FROM Contact",
        )
        query_op_mock.return_value.query.assert_called_once_with()
        task._import_results.assert_called_once_with(
            MappingStep(sf_object="Contact"), query_op_mock.return_value
        )

    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run_query__no_results(self, query_op_mock):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )
        task._import_results = mock.Mock()
        query_op_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.SUCCESS, [], 0, 0
        )

        task._run_query("SELECT Id FROM Contact", MappingStep(sf_object="Contact"))

        query_op_mock.assert_called_once_with(
            sobject="Contact",
            fields=["Id"],
            api=DataApi.SMART,
            api_options={},
            context=task,
            query="SELECT Id FROM Contact",
        )
        query_op_mock.return_value.query.assert_called_once_with()
        task._import_results.assert_not_called()

    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_run_query__failure(self, query_op_mock):
        task = _make_task(
            ExtractData, {"options": {"database_url": "sqlite:///", "mapping": ""}}
        )
        query_op_mock.return_value.job_result = DataOperationJobResult(
            DataOperationStatus.JOB_FAILURE, [], 1, 0
        )

        with pytest.raises(BulkDataException):
            task._run_query("SELECT Id FROM Contact", MappingStep(sf_object="Contact"))

    def test_init_options__missing_output(self):
        with pytest.raises(TaskOptionsError):
            _make_task(ExtractData, {"options": {}})

    @mock.patch("cumulusci.tasks.bulkdata.extract.log_progress")
    def test_extract_respects_key_field(self, log_mock):
        task = _make_task(
            ExtractData,
            {"options": {"database_url": "sqlite://", "mapping": "mapping.yml"}},
        )

        mapping = MappingStep(
            sf_object="Opportunity",
            table="Opportunity",
            fields={"Id": "sf_id", "Name": "Name"},
            lookups={
                "AccountId": MappingLookup(
                    table="Account", key_field="account_id", name="AccountId"
                )
            },
        )
        step = mock.Mock()
        task.session = mock.Mock()
        task.metadata = mock.MagicMock()

        with mock.patch(
            "cumulusci.tasks.bulkdata.extract.sql_bulk_insert_from_records"
        ) as sql_bulk_insert_from_records:
            task._import_results(mapping, step)

        task.session.connection.assert_called_once_with()
        step.get_results.assert_called_once_with()
        sql_bulk_insert_from_records.assert_called_once_with(
            connection=task.session.connection.return_value,
            table=task.metadata.tables["Opportunity"],
            columns=["sf_id", "Name", "account_id"],
            record_iterable=log_mock.return_value,
        )

    @responses.activate
    @mock.patch("cumulusci.tasks.bulkdata.extract.get_query_operation")
    def test_extract_memory_usage(self, step_mock):
        with TemporaryDirectory() as t:
            base_path = os.path.dirname(__file__)
            mapping_path = os.path.join(base_path, self.mapping_file_v1)
            mock_describe_calls()

            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{t}/foo.db",  # tempdir database
                        "mapping": mapping_path,
                    }
                },
            )
            task.bulk = mock.Mock()
            task.sf = mock.Mock()
            task.org_config._is_person_accounts_enabled = False

            mock_query_households = MockScalableBulkQueryOperation(
                sobject="Account",
                api_options={},
                context=task,
                query="SELECT Id FROM Account",
            )
            mock_query_contacts = MockScalableBulkQueryOperation(
                sobject="Contact",
                api_options={},
                context=task,
                query="SELECT Id, FirstName, LastName, Email, AccountId FROM Contact",
            )
            mock_query_households.results = ([str(num)] for num in range(1, 20000))
            mock_query_households.result_len = 20000
            mock_query_contacts.results = [
                ["2", "First", "Last", "test@example.com", "1"]
            ]
            mock_query_contacts.result_len = 1

            step_mock.side_effect = [mock_query_households, mock_query_contacts]

            with assert_max_memory_usage(15 * 10**6):
                task()

    @responses.activate
    def test_import_results__autopk(self, create_task_fixture):
        mock_describe_calls()
        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_vanilla)
        with temporary_dir() as d:
            task = create_task_fixture(
                ExtractData,
                {"database_url": f"sqlite:///{d}/temp.db", "mapping": mapping_path},
            )

            extracted_records = {
                # ['Name', 'Description', 'BillingStreet', 'BillingCity', 'BillingState', 'BillingPostalCode', 'BillingCountry', 'ParentId', 'ShippingStreet', 'ShippingCity', 'ShippingState', 'ShippingPostalCode', 'ShippingCountry', 'Phone', 'Fax', 'Website', 'NumberOfEmployees', 'AccountNumber', 'Site', 'Type', 'IsPersonAccount']
                "Account": [
                    ["001002", "Account1"] + [""] * 20 + [False],
                    ["001003", "Account2"] + [""] * 20 + [False],
                ],
                # Salutation|FirstName|LastName|Email|Phone|MobilePhone|OtherPhone|HomePhone|Title|Birthdate|MailingStreet|MailingCity|MailingState|MailingPostalCode|MailingCountry|AccountId
                "Contact": [
                    ["002001", "", "Bob 1", "Barker 2"] + [""] * 4,
                    ["002002", "", "Sam 2", "Smith 2"] + [""] * 4,
                ],  # id|Name|StageName|CloseDate|Amount|AccountId|ContactId|record_type
                "Opportunity": [
                    [
                        "0003001",
                        "Dickenson Mobile Generators",
                        "Qualification",
                        "2020-07-18",
                        "15000.0",
                        "001003",
                        "002001",
                    ]
                ],
            }
            with mock_extract_jobs(task, extracted_records), mock_salesforce_client(
                task
            ):
                task()
            with create_engine(task.options["database_url"]).connect() as conn:
                output_accounts = list(conn.execute("select * from Account"))
                assert output_accounts[0][0:2] == ("Account-1", "Account1")
                assert output_accounts[1][0:2] == ("Account-2", "Account2")
                assert len(output_accounts) == 2
                output_opportunities = list(conn.execute("select * from Opportunity"))

                assert output_opportunities[0].AccountId == "Account-2"
                assert output_opportunities[0].ContactId == "Contact-1"

    def test_run_soql_filter(self):
        """This test case is to verify when soql_filter is specified with valid filter in the mapping yml"""

        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        mock_describe_calls()
        with temporary_dir() as d:
            tmp_db_path = os.path.join(d, "testdata.db")

            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{tmp_db_path}",
                        "mapping": mapping_path,
                    }
                },
            )

            mapping = MappingStep(
                sf_object="Contact",
                fields={"Id": "Id", "Name": "Name"},
                record_type="Business",
                soql_filter="Name = 'John Doe'",
            )

            soql = task._soql_for_mapping(mapping)
            assert (
                "WHERE RecordType.DeveloperName = 'Business' AND Name = 'John Doe'"
                in soql
            ), "Filter should be applied on Name and DeveloperName"

    def test_run_soql_filter_where_specified(self):
        """This test case is to verify when soql_filter is specified with mapping yml with WHERE keyword in it"""

        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        mock_describe_calls()
        with temporary_dir() as d:
            tmp_db_path = os.path.join(d, "testdata.db")

            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{tmp_db_path}",
                        "mapping": mapping_path,
                    }
                },
            )

            mapping = MappingStep(
                sf_object="Contact",
                fields={"Id": "Id", "Name": "Name"},
                record_type="Business",
                soql_filter=" wHeRe Name = 'John Doe'",
            )

            soql = task._soql_for_mapping(mapping)
            assert (
                "WHERE RecordType.DeveloperName = 'Business' AND Name = 'John Doe'"
                in soql
            ), "Filter should be applied on Name and RecordType"

    def test_run_soql_filter_no_record_type(self):
        """This test case is to verify when soql_filter is specified in mapping yml file but no record_type"""

        base_path = os.path.dirname(__file__)
        mapping_path = os.path.join(base_path, self.mapping_file_v1)
        mock_describe_calls()
        with temporary_dir() as d:
            tmp_db_path = os.path.join(d, "testdata.db")

            task = _make_task(
                ExtractData,
                {
                    "options": {
                        "database_url": f"sqlite:///{tmp_db_path}",
                        "mapping": mapping_path,
                    }
                },
            )

            mapping = MappingStep(
                sf_object="Contact",
                fields={"Id": "Id", "Name": "Name"},
                soql_filter=" wHeRe Name = 'John Doe'",
            )

            soql = task._soql_for_mapping(mapping)
            assert (
                "WHERE Name = 'John Doe'" in soql
            ), "filter should be applied just on name"
            assert (
                "DeveloperName" not in soql
            ), "DeveloperName should not appear in the soql query as it is missing in mapping"



import io
import os
from unittest import mock

from cumulusci.tasks.salesforce import SOQLQuery
from cumulusci.utils import temporary_dir

from .util import create_task


def test_run_task():
    with temporary_dir() as path:
        task = create_task(
            SOQLQuery,
            {
                "object": "Account",
                "query": "SELECT Id FROM Account",
                "result_file": "results.csv",
            },
        )
        task.bulk = mock.Mock()
        task.bulk.get_all_results_for_query_batch.return_value = iter(
            [io.BytesIO(b"No results found.")]
        )
        task._run_task()

        task.bulk.query.assert_called_once()
        task.bulk.wait_for_batch.assert_called_once()
        task.bulk.close_job.assert_called_once()
        assert os.path.exists(os.path.join(path, "results.csv"))



import contextlib
import io
import os
import shutil
import sys
import tempfile
from pathlib import Path
from unittest import mock

import click
import pkg_resources
import pytest
from requests.exceptions import ConnectionError
from rich.console import Console

import cumulusci
from cumulusci.cli import cci
from cumulusci.cli.tests.utils import run_click_command
from cumulusci.cli.utils import get_installed_version
from cumulusci.core.config import BaseProjectConfig
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.utils import temporary_dir

MagicMock = mock.MagicMock()
CONSOLE = mock.Mock()


@pytest.fixture(autouse=True)
def env_config():
    config = {
        "global_tempdir": tempfile.gettempdir(),
        "tempdir": tempfile.mkdtemp(),
        "environ_mock": mock.patch.dict(
            os.environ, {"HOME": tempfile.mkdtemp(), "CUMULUSCI_KEY": ""}
        ),
    }
    # setup
    config["environ_mock"].start()
    assert config["global_tempdir"] in os.environ["HOME"]
    yield config
    # tear down
    assert config["global_tempdir"] in os.environ["HOME"]
    config["environ_mock"].stop()
    shutil.rmtree(config["tempdir"])


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.cli")
def test_main(
    cli,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
):
    get_tempfile_logger.return_value = mock.Mock(), "tempfile.log"
    cci.main()

    check_latest_version.assert_called_once()
    init_logger.assert_called_once()
    CliRuntime.assert_called_once()
    cli.assert_called_once()
    tee.assert_called_once()


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.cli")
@mock.patch("pdb.post_mortem")
@mock.patch("sys.exit")
def test_main__debug(
    sys_exit,
    post_mortem,
    cli,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
):
    cli.side_effect = Exception
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")

    cci.main(["cci", "--debug"])

    check_latest_version.assert_called_once()
    init_logger.assert_called_once_with(debug=True)
    CliRuntime.assert_called_once()
    cli.assert_called_once()
    post_mortem.assert_called_once()
    sys_exit.assert_called_once_with(1)
    get_tempfile_logger.assert_called_once()
    tee.assert_called_once()


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.cli")
@mock.patch("pdb.post_mortem")
def test_main__cci_show_stacktraces(
    post_mortem,
    cli,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
    capsys,
):
    runtime = mock.Mock()
    runtime.universal_config.cli__show_stacktraces = True
    CliRuntime.return_value = runtime
    cli.side_effect = Exception
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")

    with pytest.raises(SystemExit):
        cci.main(["cci"])

    check_latest_version.assert_called_once()
    init_logger.assert_called_once_with(debug=False)
    CliRuntime.assert_called_once()
    cli.assert_called_once()
    post_mortem.assert_not_called()
    captured = capsys.readouterr()
    assert "Traceback (most recent call last)" in captured.err


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.cli")
@mock.patch("sys.exit")
def test_main__abort(
    sys_exit, cli, init_logger, get_tempfile_logger, tee_stdout_stderr
):
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")
    cli.side_effect = click.Abort
    cci.main(["cci"])
    cli.assert_called_once()
    sys_exit.assert_called_once_with(1)


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.cli")
@mock.patch("pdb.post_mortem")
@mock.patch("sys.exit")
def test_main__error(
    sys_exit,
    post_mortem,
    cli,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
):
    runtime = mock.Mock()
    runtime.universal_config.cli__show_stacktraces = False
    CliRuntime.return_value = runtime

    cli.side_effect = Exception
    get_tempfile_logger.return_value = mock.Mock(), "tempfile.log"

    cci.main(["cci", "org", "info"])

    check_latest_version.assert_called_once()
    init_logger.assert_called_once_with(debug=False)
    CliRuntime.assert_called_once()
    cli.assert_called_once()
    post_mortem.call_count == 0
    sys_exit.assert_called_once_with(1)
    get_tempfile_logger.assert_called_once()
    tee.assert_called_once()

    os.remove("tempfile.log")


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.CliRuntime")
def test_main__CliRuntime_error(CliRuntime, get_tempfile_logger, tee):
    CliRuntime.side_effect = CumulusCIException("something happened")
    get_tempfile_logger.return_value = mock.Mock(), "tempfile.log"

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        with mock.patch("sys.exit") as sys_exit:
            sys_exit.side_effect = SystemExit  # emulate real sys.exit
            with pytest.raises(SystemExit):
                cci.main(["cci", "org", "info"])

    assert "something happened" in stderr.getvalue()

    tempfile = Path("tempfile.log")
    tempfile.unlink()


@mock.patch("cumulusci.cli.cci.init_logger")  # side effects break other tests
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("sys.exit", MagicMock())
def test_handle_org_name(
    CliRuntime, tee_stdout_stderr, get_tempfile_logger, init_logger
):

    # get_tempfile_logger doesn't clean up after itself which breaks other tests
    get_tempfile_logger.return_value = mock.Mock(), ""

    with contextlib.redirect_stdout(io.StringIO()) as stdout:
        cci.main(["cci", "org", "default", "xyzzy"])
    assert "xyzzy is now the default org" in stdout.getvalue()

    with contextlib.redirect_stdout(io.StringIO()) as stdout:
        cci.main(["cci", "org", "default", "--org", "xyzzy2"])
    assert "xyzzy2 is now the default org" in stdout.getvalue()

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.main(["cci", "org", "default", "xyzzy1", "--org", "xyzzy2"])
    assert "not both" in stderr.getvalue()

    CliRuntime().keychain.get_default_org.return_value = ("xyzzy3", None)

    # cci org remove should really need an attached org
    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.main(["cci", "org", "remove"])
    assert "Please specify ORGNAME or --org ORGNAME" in stderr.getvalue()


@mock.patch("cumulusci.cli.cci.init_logger")  # side effects break other tests
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("sys.exit")
@mock.patch("cumulusci.cli.cci.CliRuntime")
def test_cci_org_default__no_orgname(
    CliRuntime, exit, tee_stdout_stderr, get_tempfile_logger, init_logger
):
    # get_tempfile_logger doesn't clean up after itself which breaks other tests
    get_tempfile_logger.return_value = mock.Mock(), ""

    CliRuntime().keychain.get_default_org.return_value = ("xyzzy4", None)
    with contextlib.redirect_stdout(io.StringIO()) as stdout:
        cci.main(["cci", "org", "default"])
    assert "xyzzy4 is the default org" in stdout.getvalue()

    CliRuntime().keychain.get_default_org.return_value = (None, None)
    with contextlib.redirect_stdout(io.StringIO()) as stdout:
        cci.main(["cci", "org", "default"])
    assert "There is no default org" in stdout.getvalue()


DEPLOY_CLASS_PATH = f"cumulusci.tasks.salesforce.Deploy{'.Deploy' if sys.version_info >= (3, 11) else ''}"


@mock.patch("cumulusci.cli.cci.init_logger", mock.Mock())
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr", mock.MagicMock())
@mock.patch(f"{DEPLOY_CLASS_PATH}.__call__", mock.Mock())
@mock.patch("sys.exit", mock.Mock())
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch(f"{DEPLOY_CLASS_PATH}.__init__")
def test_cci_run_task_options__with_dash(
    Deploy,
    CliRuntime,
    get_tempfile_logger,
):
    # get_tempfile_logger doesn't clean up after itself which breaks other tests
    Deploy.return_value = None
    get_tempfile_logger.return_value = mock.Mock(), ""
    CliRuntime.return_value = runtime = mock.Mock()
    runtime.get_org.return_value = ("test", mock.Mock())
    runtime.project_config = BaseProjectConfig(
        runtime.universal_config,
        {
            "project": {"name": "Test"},
            "tasks": {"deploy": {"class_path": "cumulusci.tasks.salesforce.Deploy"}},
        },
    )

    cci.main(
        ["cci", "task", "run", "deploy", "--path", "x", "--clean-meta-xml", "False"]
    )
    task_config = Deploy.mock_calls[0][1][1]
    assert "clean_meta_xml" in task_config.options


@mock.patch("cumulusci.cli.cci.init_logger", mock.Mock())
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr", mock.MagicMock())
@mock.patch(f"{DEPLOY_CLASS_PATH}.__call__", mock.Mock())
@mock.patch("sys.exit", mock.Mock())
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch(f"{DEPLOY_CLASS_PATH}.__init__")
def test_cci_run_task_options__old_style_with_dash(
    Deploy,
    CliRuntime,
    get_tempfile_logger,
):
    # get_tempfile_logger doesn't clean up after itself which breaks other tests
    Deploy.return_value = None
    get_tempfile_logger.return_value = mock.Mock(), ""
    CliRuntime.return_value = runtime = mock.Mock()
    runtime.get_org.return_value = ("test", mock.Mock())
    runtime.project_config = BaseProjectConfig(
        runtime.universal_config,
        {
            "project": {"name": "Test"},
            "tasks": {"deploy": {"class_path": "cumulusci.tasks.salesforce.Deploy"}},
        },
    )

    cci.main(
        [
            "cci",
            "task",
            "run",
            "deploy",
            "--path",
            "x",
            "-o",
            "clean-meta-xml",
            "False",
        ]
    )
    task_config = Deploy.mock_calls[0][1][1]
    assert "clean_meta_xml" in task_config.options


@mock.patch("cumulusci.cli.cci.open")
@mock.patch("cumulusci.cli.cci.traceback")
def test_handle_exception(traceback, cci_open):
    console = mock.Mock()
    Console.return_value = console
    error_message = "foo"
    cci_open.__enter__.return_value = mock.Mock()

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.handle_exception(error_message, False, "logfile.path")

    stderr = stderr.getvalue()
    assert f"Error: {error_message}" in stderr
    assert cci.SUGGEST_ERROR_COMMAND in stderr
    traceback.print_exc.assert_called_once()


@mock.patch("cumulusci.cli.cci.open")
def test_handle_exception__error_cmd(cci_open):
    """Ensure we don't write to logfiles when running `cci error ...` commands."""
    error_message = "foo"
    logfile_path = None

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.handle_exception(error_message, False, logfile_path)

    stderr = stderr.getvalue()
    assert f"Error: {error_message}" in stderr
    assert cci.SUGGEST_ERROR_COMMAND in stderr
    cci_open.assert_not_called()


@mock.patch("cumulusci.cli.cci.open")
@mock.patch("cumulusci.cli.cci.traceback")
def test_handle_click_exception(traceback, cci_open):
    cci_open.__enter__.return_value = mock.Mock()

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.handle_exception(click.ClickException("[oops]"), False, "file.path")

    stderr = stderr.getvalue()
    assert "Error: [oops]" in stderr
    traceback.assert_not_called()


@mock.patch("cumulusci.cli.cci.open")
def test_handle_connection_exception(cci_open):
    cci_open.__enter__.return_value = mock.Mock()

    with contextlib.redirect_stderr(io.StringIO()) as stderr:
        cci.handle_exception(ConnectionError(), False, "file.log")

    stderr = stderr.getvalue()
    assert "We encountered an error with your internet connection." in stderr


def test_cli():
    run_click_command(cci.cli)


@mock.patch(
    "cumulusci.cli.cci.get_latest_final_version",
    mock.Mock(return_value=pkg_resources.parse_version("100")),
)
def test_version(capsys):
    run_click_command(cci.version)
    console_output = capsys.readouterr().out
    assert f"CumulusCI version: {cumulusci.__version__}" in console_output
    assert "There is a newer version of CumulusCI available" in console_output


@mock.patch(
    "cumulusci.cli.cci.get_latest_final_version",
    mock.Mock(return_value=pkg_resources.parse_version("1")),
)
def test_version__latest(capsys):
    run_click_command(cci.version)
    console_output = capsys.readouterr().out
    assert "You have the latest version of CumulusCI" in console_output


@mock.patch("cumulusci.cli.cci.warn_if_no_long_paths")
@mock.patch("cumulusci.cli.cci.get_latest_final_version", get_installed_version)
def test_version__win_path_warning(warn_if):
    run_click_command(cci.version)
    warn_if.assert_called_once()


@mock.patch("code.interact")
def test_shell(interact):
    run_click_command(cci.shell)
    interact.assert_called_once()
    assert "config" in interact.call_args[1]["local"]
    assert "runtime" in interact.call_args[1]["local"]


@mock.patch("runpy.run_path")
def test_shell_script(runpy):
    run_click_command(cci.shell, script="foo.py")
    runpy.assert_called_once()
    assert "config" in runpy.call_args[1]["init_globals"]
    assert "runtime" in runpy.call_args[1]["init_globals"]
    assert runpy.call_args[0][0] == "foo.py", runpy.call_args[0]


@mock.patch("builtins.print")
def test_shell_code(print):
    run_click_command(cci.shell, python="print(config, runtime)")
    print.assert_called_once()


@mock.patch("cumulusci.cli.cci.print")
def test_shell_mutually_exclusive_args(print):
    with pytest.raises(Exception) as e:
        run_click_command(cci.shell, script="foo.py", python="print(config, runtime)")
    assert "Cannot specify both" in e.value.message


@mock.patch("code.interact")
def test_shell__no_project(interact):
    with temporary_dir():
        run_click_command(cci.shell)
        interact.assert_called_once()


def test_cover_command_groups():
    run_click_command(cci.project)
    run_click_command(cci.org)
    run_click_command(cci.task)
    run_click_command(cci.flow)
    run_click_command(cci.service)
    # no assertion; this test is for coverage of empty methods


@mock.patch(
    "cumulusci.cli.runtime.CliRuntime.get_org",
    lambda *args, **kwargs: (MagicMock(), MagicMock()),
)
@mock.patch("cumulusci.core.runtime.BaseCumulusCI._load_keychain", MagicMock())
@mock.patch("pdb.post_mortem", MagicMock())
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr", MagicMock())
@mock.patch("cumulusci.cli.cci.init_logger", MagicMock())
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
def test_run_task_debug(get_tempfile_logger):
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")

    gipnew = "cumulusci.tasks.preflight.packages.GetInstalledPackages._run_task"
    with mock.patch(gipnew, mock_validate_debug(False)):
        cci.main(["cci", "task", "run", "get_installed_packages"])
    with mock.patch(gipnew, mock_validate_debug(True)):
        cci.main(["cci", "task", "run", "get_installed_packages", "--debug"])


@mock.patch(
    "cumulusci.cli.runtime.CliRuntime.get_org",
    lambda *args, **kwargs: (MagicMock(), MagicMock()),
)
@mock.patch("cumulusci.core.runtime.BaseCumulusCI._load_keychain", MagicMock())
@mock.patch("pdb.post_mortem", MagicMock())
@mock.patch("cumulusci.cli.cci.tee_stdout_stderr", MagicMock())
@mock.patch("cumulusci.cli.cci.init_logger", MagicMock())
@mock.patch("cumulusci.tasks.robotframework.RobotLibDoc", MagicMock())
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
def test_run_flow_debug(get_tempfile_logger):
    get_tempfile_logger.return_value = (mock.Mock(), "tempfile.log")
    rtd = "cumulusci.tasks.robotframework.RobotTestDoc._run_task"

    with mock.patch(rtd, mock_validate_debug(False)):
        cci.main(["cci", "flow", "run", "robot_docs"])
    with mock.patch(rtd, mock_validate_debug(True)):
        cci.main(["cci", "flow", "run", "robot_docs", "--debug"])


def mock_validate_debug(value):
    def _run_task(self, *args, **kwargs):
        assert bool(self.debug_mode) == bool(value)

    return _run_task


@mock.patch("cumulusci.cli.cci.tee_stdout_stderr")
@mock.patch("cumulusci.cli.cci.get_tempfile_logger")
@mock.patch("cumulusci.cli.cci.init_logger")
@mock.patch("cumulusci.cli.cci.check_latest_version")
@mock.patch("cumulusci.cli.cci.CliRuntime")
@mock.patch("cumulusci.cli.cci.show_version_info")
def test_dash_dash_version(
    show_version_info,
    CliRuntime,
    check_latest_version,
    init_logger,
    get_tempfile_logger,
    tee,
):
    get_tempfile_logger.return_value = mock.Mock(), "tempfile.log"
    cci.main(["cci", "--help"])
    assert len(show_version_info.mock_calls) == 0

    cci.main(["cci", "version"])
    assert len(show_version_info.mock_calls) == 1

    cci.main(["cci", "--version"])
    assert len(show_version_info.mock_calls) == 2



import base64
import json
import pathlib
import re

from simple_salesforce.exceptions import SalesforceMalformedRequest

from cumulusci.core.exceptions import CumulusCIException
from cumulusci.tasks.salesforce import BaseSalesforceApiTask


def join_errors(e: SalesforceMalformedRequest) -> str:
    return "; ".join([error.get("message", "Unknown.") for error in e.content])


class UploadProfilePhoto(BaseSalesforceApiTask):
    task_docs = """
Uploads a profile photo for a specified or default User.

Examples
--------

Upload a profile photo for the default user.

.. code-block:: yaml

    tasks:
        upload_profile_photo_default:
            group: Internal storytelling data
            class_path: cumulusci.tasks.salesforce.users.UploadProfilePhoto
            description: Uploads a profile photo for the default user.
            options:
                photo: storytelling/photos/default.png

Upload a profile photo for a user whose Alias equals ``grace`` or ``walker``, is active, and created today.

.. code-block:: yaml

    tasks:
        upload_profile_photo_grace:
            group: Internal storytelling data
            class_path: cumulusci.tasks.salesforce.users.UploadProfilePhoto
            description: Uploads a profile photo for Grace.
            options:
                photo: storytelling/photos/grace.png
                where: (Alias = 'grace' OR Alias = 'walker') AND IsActive = true AND CreatedDate = TODAY
    """

    task_options = {
        "photo": {"description": "Path to user's profile photo.", "required": True},
        "where": {
            "description": """WHERE clause used querying for which User to upload the profile photo for.

* No need to prefix with ``WHERE``

* The SOQL query must return one and only one User record.

* If no "where" is supplied, uploads the photo for the org's default User.

""",
            "required": False,
        },
    }

    def _raise_cumulusci_exception(self, e: SalesforceMalformedRequest) -> None:
        raise CumulusCIException(join_errors(e))

    def _get_user_id_by_query(self, where: str) -> str:
        # Query for the User removing a "WHERE " prefix from where if exists.
        query = "SELECT Id FROM User WHERE {} LIMIT 2".format(
            re.sub(r"^WHERE ", "", where, flags=re.I)
        )
        self.logger.info(f"Querying User: {query}")

        user_ids = []
        try:
            for record in self.sf.query_all(query)["records"]:
                user_ids.append(record["Id"])
        except SalesforceMalformedRequest as e:
            # Raise an easier to digest exception.
            self._raise_cumulusci_exception(e)

        # Validate only 1 User found.
        if len(user_ids) < 1:
            raise CumulusCIException("No Users found.")
        if 1 < len(user_ids):
            raise CumulusCIException(
                "More than one User found (at least 2): {}".format(", ".join(user_ids))
            )

        # Log and return User ID.
        self.logger.info(f"Uploading profile photo for the User with ID {user_ids[0]}")
        return user_ids[0]

    def _get_default_user_id(self) -> str:
        user_id = self.sf.restful("")["identity"][-18:]
        self.logger.info(
            f"Uploading profile photo for the default User with ID {user_id}"
        )
        return user_id

    def _insert_content_document(self, photo_path) -> str:
        path = pathlib.Path(photo_path)

        if not path.exists():
            raise CumulusCIException(f"No photo found at {path}")

        self.logger.info(f"Setting user photo to {path}")
        result = self.sf.ContentVersion.create(
            {
                "PathOnClient": path.name,
                "Title": path.stem,
                "VersionData": base64.b64encode(path.read_bytes()).decode("utf-8"),
            }
        )
        if not result["success"]:
            raise CumulusCIException(
                "Failed to create photo ContentVersion: {}".format(result["errors"])
            )
        content_version_id = result["id"]

        # Query the ContentDocumentId for our created record.
        content_document_id = self.sf.query(
            f"SELECT Id, ContentDocumentId FROM ContentVersion WHERE Id = '{content_version_id}'"
        )["records"][0]["ContentDocumentId"]

        self.logger.info(
            f"Uploaded profile photo ContentDocument {content_document_id}"
        )

        return content_document_id

    def _delete_content_document(self, content_document_id: str):
        self.sf.ContentDocument.delete(content_document_id)

    def _assign_user_profile_photo(self, user_id: str, content_document_id: str):
        # Call the Connect API to set our user photo.
        try:
            self.sf.restful(
                f"connect/user-profiles/{user_id}/photo",
                data=json.dumps({"fileId": content_document_id}),
                method="POST",
            )
        except SalesforceMalformedRequest as e:
            # Rollback ContentDocument, and raise an easier to digest exception.
            self.logger.error(
                "An error occured assigning the ContentDocument as the users's profile photo."
            )
            self.logger.error(f"Deleting ContentDocument {content_document_id}")
            self._delete_content_document(content_document_id)
            self._raise_cumulusci_exception(e)

    def _run_task(self):
        user_id = (
            self._get_user_id_by_query(self.options["where"])
            if self.options.get("where")
            else self._get_default_user_id()
        )

        content_document_id = self._insert_content_document(self.options["photo"])

        self._assign_user_profile_photo(user_id, content_document_id)



import contextlib
import os
import re
import sys
import time
from collections import defaultdict

import click
import pkg_resources
import requests
from rich.console import Console

from cumulusci import __version__
from cumulusci.core.config import UniversalConfig
from cumulusci.utils import get_cci_upgrade_command
from cumulusci.utils.http.requests_utils import safe_json_from_response

LOWEST_SUPPORTED_VERSION = (3, 8, 0)
WIN_LONG_PATH_WARNING = """
WARNING: Long path support is not enabled. This can lead to errors with some
tasks. Your administrator will need to activate the "Enable Win32 long paths"
group policy, or set LongPathsEnabled to 1 in the registry key
HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Control\\FileSystem.
"""


def group_items(items):
    """Given a list of dicts with 'group' keys,
    returns those items in lists categorized group"""
    groups = defaultdict(list)
    for item in items:
        group_name = item["group"] or "Other"
        groups[group_name].append([item["name"], item["description"]])

    return groups


@contextlib.contextmanager
def timestamp_file():
    """Opens a file for tracking the time of the last version check"""

    config_dir = UniversalConfig.default_cumulusci_dir()
    timestamp_file = os.path.join(config_dir, "cumulus_timestamp")

    try:
        with open(timestamp_file, "r+") as f:
            yield f
    except IOError:  # file does not exist
        with open(timestamp_file, "w+") as f:
            yield f


FINAL_VERSION_RE = re.compile(r"^[\d\.]+$")


def is_final_release(version: str) -> bool:
    """Returns bool whether version string should be considered a final release.

    cumulusci versions are considered final if they contain only digits and periods.
    e.g. 1.0.1 is final but 2.0b1 and 2.0.dev0 are not.
    """
    return bool(FINAL_VERSION_RE.match(version))


def get_latest_final_version():
    """return the latest version of cumulusci in pypi, be defensive"""
    # use the pypi json api https://wiki.python.org/moin/PyPIJSON
    res = safe_json_from_response(
        requests.get("https://pypi.org/pypi/cumulusci/json", timeout=5)
    )
    with timestamp_file() as f:
        f.write(str(time.time()))
    versions = []
    for versionstring in res["releases"].keys():
        if not is_final_release(versionstring):
            continue
        versions.append(pkg_resources.parse_version(versionstring))
    versions.sort(reverse=True)
    return versions[0]


def check_latest_version():
    """checks for the latest version of cumulusci from pypi, max once per hour"""
    check = True

    with timestamp_file() as f:
        timestamp = float(f.read() or 0)
    delta = time.time() - timestamp
    check = delta > 3600

    if check:
        try:
            latest_version = get_latest_final_version()
        except requests.exceptions.RequestException as e:
            click.echo("Error checking cci version:", err=True)
            click.echo(str(e), err=True)
            return

        result = latest_version > get_installed_version()
        if result:
            click.echo(
                f"""An update to CumulusCI is available. To install the update, run this command: {get_cci_upgrade_command()}""",
                err=True,
            )

        if sys.version_info < LOWEST_SUPPORTED_VERSION:
            click.echo(
                "Sorry! Your Python version is not supported. Please upgrade to Python 3.9.",
                err=True,
            )


def get_installed_version():
    """returns the version name (e.g. 2.0.0b58) that is installed"""
    return pkg_resources.parse_version(__version__)


def win32_long_paths_enabled() -> bool:
    """Boolean indicating whether long paths are available on Windows systems.

    Reads the Windows Registry the running platform. Throws ModuleNotFoundError
    if run on non-Windows platforms.
    """
    # Only present on windows, so import it here instead
    import winreg

    access_registry = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
    access_key = winreg.OpenKey(
        access_registry, r"SYSTEM\CurrentControlSet\Control\FileSystem"
    )

    is_enabled, _ = winreg.QueryValueEx(access_key, "LongPathsEnabled")

    return is_enabled == 1


def warn_if_no_long_paths(console: Console = Console()) -> None:
    """Print a warning to the user if long paths are not enabled."""
    if sys.platform.startswith("win") and not win32_long_paths_enabled():
        console.print(WIN_LONG_PATH_WARNING)



# -*- coding: utf-8 -*-
import json
import os
import pathlib
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

import pytest
import responses
import yaml
from github3.exceptions import NotFoundError
from simple_salesforce.exceptions import SalesforceError

from cumulusci.core.config import (
    BaseConfig,
    BaseProjectConfig,
    BaseTaskFlowConfig,
    OrgConfig,
    ServiceConfig,
    UniversalConfig,
)
from cumulusci.core.config.org_config import VersionInfo
from cumulusci.core.dependencies.dependencies import (
    PackageNamespaceVersionDependency,
    PackageVersionIdDependency,
)
from cumulusci.core.exceptions import (
    ConfigError,
    CumulusCIException,
    DependencyResolutionError,
    FlowNotFoundError,
    GithubException,
    KeychainNotFound,
    NamespaceNotFoundError,
    ServiceNotConfigured,
    TaskNotFoundError,
)
from cumulusci.core.keychain.base_project_keychain import (
    DEFAULT_CONNECTED_APP,
    BaseProjectKeychain,
)
from cumulusci.core.source import LocalFolderSource
from cumulusci.tests.util import CURRENT_SF_API_VERSION, DummyKeychain
from cumulusci.utils import temporary_dir, touch
from cumulusci.utils.version_strings import StrictVersion
from cumulusci.utils.yaml.cumulusci_yml import GitHubSourceModel, LocalFolderSourceModel


class FakeConfig(BaseConfig):
    foo: dict


class TestBaseConfig:
    def test_getattr_toplevel_key(self):
        config = FakeConfig()
        config.config = {"foo": "bar"}
        assert config.foo == "bar"

    def test_getattr_toplevel_key_missing(self):
        config = BaseConfig()
        config.config = {}
        with mock.patch(
            "cumulusci.core.config.base_config.STRICT_GETATTR", False
        ), pytest.warns(DeprecationWarning, match="foo"):
            assert config.foo is None
        with mock.patch(
            "cumulusci.core.config.base_config.STRICT_GETATTR", True
        ), pytest.raises(AssertionError):
            assert config.foo is None

    def test_getattr_child_key(self):
        config = FakeConfig()
        config.config = {"foo": {"bar": "baz"}}
        assert config.foo__bar == "baz"

    def test_strict_getattr(self):
        config = FakeConfig()
        config.config = {"foo": {"bar": "baz"}}
        with mock.patch(
            "cumulusci.core.config.base_config.STRICT_GETATTR", "True"
        ), mock.patch("warnings.warn"), pytest.raises(AssertionError):
            print(config.jfiesojfieoj)

    def test_getattr_child_parent_key_missing(self):
        config = FakeConfig()
        config.config = {}
        assert config.foo__bar is None

    def test_getattr_child_key_missing(self):
        config = FakeConfig()
        config.config = {"foo": {}}
        assert config.foo__bar is None

    def test_getattr_default_toplevel(self):
        config = FakeConfig()
        config.config = {"foo": "bar"}
        config.defaults = {"foo": "default"}
        assert config.foo == "bar"

    def test_getattr_default_toplevel_missing_default(self):
        config = FakeConfig()
        config.config = {"foo": "bar"}
        config.defaults = {}
        assert config.foo == "bar"

    def test_getattr_default_toplevel_missing_config(self):
        config = FakeConfig()
        config.config = {}
        config.defaults = {"foo": "default"}
        assert config.foo == "default"

    def test_getattr_default_child(self):
        config = FakeConfig()
        config.config = {"foo": {"bar": "baz"}}
        config.defaults = {"foo__bar": "default"}
        assert config.foo__bar == "baz"

    def test_getattr_default_child_missing_default(self):
        config = FakeConfig()
        config.config = {"foo": {"bar": "baz"}}
        config.defaults = {}
        assert config.foo__bar == "baz"

    def test_getattr_default_child_missing_config(self):
        config = FakeConfig()
        config.config = {}
        config.defaults = {"foo__bar": "default"}
        assert config.foo__bar == "default"


class DummyContents(object):
    def __init__(self, content):
        self.decoded = content


class DummyResponse(object):
    def __init__(self, content, status_code):
        self.content = content
        self.status_code = status_code


class DummyRepository(mock.Mock):
    default_branch = "main"
    _api = "http://"

    def __init__(self, owner, name, contents, releases=None, commits=None, **kwargs):
        """Passing kwargs to workaround python/cpython#83759"""
        super().__init__()
        self.owner = owner
        self.name = name
        self.html_url = f"https://github.com/{owner}/{name}"
        self.clone_url = self.html_url
        self.session = mock.MagicMock()
        self._contents = contents
        self._releases = releases or []
        self._commits = commits or []

    def file_contents(self, path, **kw):
        try:
            return self._contents[path]
        except KeyError:
            raise AssertionError(f"Accessed unexpected file: {path}")

    def directory_contents(self, path, **kw):
        try:
            return self._contents[path]
        except KeyError:
            raise NotFoundError(
                DummyResponse(f"Accessed unexpected directory: {path}", 404)
            )

    def _build_url(self, *args, **kw):
        return self._api

    def releases(self):
        return iter(self._releases)

    def latest_release(self):
        for release in self._releases:
            if release.tag_name.startswith("release/"):
                return release
        raise NotFoundError(DummyResponse("", 404))

    def release_from_tag(self, tag_name):
        for release in self._releases:
            if release.tag_name == tag_name:
                return release
        raise NotFoundError(DummyResponse("", 404))

    def branch(self, name):
        branch = mock.Mock()
        branch.commit.sha = "commit_sha"
        branch.name = name
        return branch

    def tag(self, sha):
        tag = mock.Mock()
        tag.object.sha = "tag_sha"
        tag.message = ""
        return tag

    def ref(self, s):
        ref = mock.Mock()
        ref.object.sha = "ref_sha"
        return ref

    def commit(self, c):
        if c in self._commits:
            return self._commits[c]

        raise NotFoundError(DummyResponse("", 404))


class DummyRelease(object):
    def __init__(self, tag_name, name=None):
        self.tag_name = tag_name
        self.name = name


class DummyGithub(object):
    def __init__(self, repositories):
        self.repositories = repositories

    def repository(self, owner, name):
        try:
            return self.repositories[name]
        except KeyError:
            raise AssertionError(f"Unexpected repository: {name}")


class TestBaseProjectConfig:
    maxDiff = None

    def _make_github(self):
        CUMULUSCI_TEST_REPO = DummyRepository(
            "SFDO-Tooling",
            "CumulusCI-Test",
            {
                "cumulusci.yml": DummyContents(
                    b"""
        project:
            name: CumulusCI-Test
            package:
                name: CumulusCI-Test
                namespace: ccitest
            git:
                repo_url: https://github.com/SFDO-Tooling/CumulusCI-Test
            dependencies:
                - github: https://github.com/SFDO-Tooling/CumulusCI-Test-Dep
        """
                ),
                "unpackaged/pre": {"pre": {}, "skip": {}},
                "src": {"src": ""},
                "unpackaged/post": {"post": {}, "skip": {}},
            },
        )

        CUMULUSCI_TEST_DEP_REPO = DummyRepository(
            "SFDO-Tooling",
            "CumulusCI-Test-Dep",
            {
                "cumulusci.yml": DummyContents(
                    b"""
        project:
            name: CumulusCI-Test-Dep
            package:
                name: CumulusCI-Test-Dep
                namespace: ccitestdep
            git:
                repo_url: https://github.com/SFDO-Tooling/CumulusCI-Test-Dep
        """
                ),
                "unpackaged/pre": {},
                "src": {},
                "unpackaged/post": {},
            },
            [DummyRelease("release/2.0", "2.0")],
        )

        CUMULUSCI_REPO = DummyRepository(
            "SFDO-Tooling",
            "CumulusCI",
            {},
            [
                DummyRelease("release/1.1", "1.1"),
                DummyRelease("beta-wrongprefix", "wrong"),
                DummyRelease("release/1.0", "1.0"),
                DummyRelease("beta/1.0-Beta_2", "1.0 (Beta 2)"),
                DummyRelease("beta/1.0-Beta_1", "1.0 (Beta 1)"),
            ],
        )

        return DummyGithub(
            {
                "CumulusCI": CUMULUSCI_REPO,
                "CumulusCI-Test": CUMULUSCI_TEST_REPO,
                "CumulusCI-Test-Dep": CUMULUSCI_TEST_DEP_REPO,
            }
        )

    def test_config_global(self):
        universal_config = UniversalConfig()
        universal_config.config_global = {}
        config = BaseProjectConfig(universal_config)
        assert universal_config.config_global is config.config_global

    def test_config_universal(self):
        universal_config = UniversalConfig()
        config = BaseProjectConfig(universal_config)
        assert universal_config.config_universal is config.config_universal

    def test_repo_info(self):
        env = {
            "CUMULUSCI_AUTO_DETECT": "1",
            "HEROKU_TEST_RUN_ID": "TEST1",
            "HEROKU_TEST_RUN_BRANCH": "main",
            "HEROKU_TEST_RUN_COMMIT_VERSION": "HEAD",
            "CUMULUSCI_REPO_BRANCH": "feature/test",
            "CUMULUSCI_REPO_COMMIT": "HEAD~1",
            "CUMULUSCI_REPO_ROOT": ".",
            "CUMULUSCI_REPO_URL": "https://github.com/SFDO-Tooling/CumulusCI-Test.git",
        }
        with mock.patch.dict(os.environ, env):
            config = BaseProjectConfig(UniversalConfig())
            result = config.repo_info
        assert {
            "ci": "heroku",
            "name": "CumulusCI-Test",
            "owner": "SFDO-Tooling",
            "branch": "feature/test",
            "commit": "HEAD~1",
            "root": ".",
            "url": "https://github.com/SFDO-Tooling/CumulusCI-Test.git",
        } == result

    def test_repo_info_missing_env(self):
        env = {
            "CUMULUSCI_AUTO_DETECT": "1",
            "HEROKU_TEST_RUN_ID": "TEST1",
            "HEROKU_TEST_RUN_BRANCH": "main",
            "HEROKU_TEST_RUN_COMMIT_VERSION": "HEAD",
            "CUMULUSCI_REPO_BRANCH": "feature/test",
            "CUMULUSCI_REPO_COMMIT": "HEAD~1",
            "CUMULUSCI_REPO_ROOT": ".",
        }
        with mock.patch.dict(os.environ, env):
            with pytest.raises(ConfigError):
                config = BaseProjectConfig(UniversalConfig())
                config.repo_info

    def test_repo_root_from_env(self):
        config = BaseProjectConfig(UniversalConfig())
        config._repo_info = {"root": "."}
        assert config.repo_root == "."

    def test_server_domain_from_repo_info(self):
        config = BaseProjectConfig(UniversalConfig())
        assert config.server_domain == "github.com"

    def test_server_domain_no_repo_root(self):
        config = BaseProjectConfig(UniversalConfig())
        with temporary_dir():
            assert config.server_domain is None

    def test_repo_name_from_repo_info(self):
        config = BaseProjectConfig(UniversalConfig())
        config._repo_info = {"name": "CumulusCI"}
        assert config.repo_name == "CumulusCI"

    def test_repo_name_no_repo_root(self):
        config = BaseProjectConfig(UniversalConfig())
        with temporary_dir():
            assert config.repo_name is None

    def test_repo_name_from_git(self):
        config = BaseProjectConfig(UniversalConfig())
        assert config.repo_name == "CumulusCI"

    def test_repo_url_from_repo_info(self):
        config = BaseProjectConfig(UniversalConfig())
        config._repo_info = {"url": "https://github.com/SFDO-Tooling/CumulusCI"}
        assert config.repo_url == "https://github.com/SFDO-Tooling/CumulusCI"

    def test_lookup_repo_branch(self):
        config = BaseProjectConfig(UniversalConfig())
        config._repo_info = {"branch": "foo-bar-baz"}
        assert config.lookup("repo_branch") == "foo-bar-baz"

    def test_repo_url_no_repo_root(self):
        config = BaseProjectConfig(UniversalConfig())
        with temporary_dir():
            assert config.repo_url is None

    @mock.patch("cumulusci.core.config.project_config.git_path")
    def test_repo_url_from_git(self, git_path):
        git_config_file = "git_config"
        git_path.return_value = git_config_file
        repo_url = "https://github.com/foo/bar.git"
        with open(git_config_file, "w") as f:
            f.writelines(['[remote "origin"]\n' f"\turl = {repo_url}"])

        config = BaseProjectConfig(UniversalConfig())
        assert repo_url == config.repo_url

        os.remove(git_config_file)

    def test_repo_owner_from_repo_info(self):
        config = BaseProjectConfig(UniversalConfig())
        config._repo_info = {"owner": "SFDO-Tooling"}
        assert config.repo_owner == "SFDO-Tooling"

    def test_repo_owner_no_repo_root(self):
        config = BaseProjectConfig(UniversalConfig())
        with temporary_dir():
            assert config.repo_owner is None

    def test_repo_branch_from_repo_info(self):
        config = BaseProjectConfig(UniversalConfig())
        config._repo_info = {"branch": "main"}
        assert config.repo_branch == "main"

    def test_repo_branch_no_repo_root(self):
        config = BaseProjectConfig(UniversalConfig())
        with temporary_dir():
            assert config.repo_branch is None

    def test_repo_commit_from_repo_info(self):
        config = BaseProjectConfig(UniversalConfig())
        config._repo_info = {"commit": "abcdef"}
        assert config.repo_commit == "abcdef"

    def test_repo_commit_no_repo_root(self):
        config = BaseProjectConfig(UniversalConfig())
        with temporary_dir():
            assert config.repo_commit is None

    def test_repo_commit_no_repo_branch(self):
        config = BaseProjectConfig(UniversalConfig())
        with temporary_dir() as d:
            os.mkdir(os.path.join(d, ".git"))
            with open(os.path.join(d, ".git", "HEAD"), "w") as f:
                f.write("abcdef")

            assert config.repo_commit == "abcdef"

    def test_repo_commit_packed_refs(self):
        config = BaseProjectConfig(UniversalConfig())
        with temporary_dir() as d:
            os.system("git init")
            with open(os.path.join(d, ".git", "HEAD"), "w") as f:
                f.write("ref: refs/heads/main\n")
            with open(os.path.join(d, ".git", "packed-refs"), "w") as f:
                f.write("# pack-refs with: peeled fully-peeled sorted\n")
                f.write("#\n")
                f.write(
                    "8ce67f4519190cd1ec9785105168e21b9599bc27 refs/remotes/origin/main\n"
                )

            assert config.repo_commit is not None

    def test_get_repo_from_url(self):
        config = BaseProjectConfig(
            UniversalConfig(),
            {},
        )

        config.get_github_api = mock.Mock()

        assert config.get_repo_from_url("https://github.com/Test/TestRepo") == (
            config.get_github_api.return_value.repository.return_value
        )
        config.get_github_api.assert_called_once_with(
            "https://github.com/Test/TestRepo"
        )
        config.get_github_api.return_value.repository.assert_called_once_with(
            "Test", "TestRepo"
        )

    def test_get_latest_tag(self):
        config = BaseProjectConfig(
            UniversalConfig(),
            {
                "project": {
                    "git": {"prefix_beta": "beta/", "prefix_release": "release/"}
                }
            },
        )
        config.get_github_api = mock.Mock(return_value=self._make_github())
        result = config.get_latest_tag()
        assert result == "release/1.1"

    def test_get_latest_tag_matching_prefix(self):
        config = BaseProjectConfig(
            UniversalConfig(),
            {"project": {"git": {"prefix_beta": "beta/", "prefix_release": "rel/"}}},
        )
        github = self._make_github()
        github.repositories["CumulusCI"]._releases.append(
            DummyRelease("rel/0.9", "0.9")
        )
        config.get_github_api = mock.Mock(return_value=github)
        result = config.get_latest_tag()
        assert result == "rel/0.9"

    def test_get_latest_tag_beta(self):
        config = BaseProjectConfig(
            UniversalConfig(),
            {
                "project": {
                    "git": {"prefix_beta": "beta/", "prefix_release": "release/"}
                }
            },
        )
        config.get_github_api = mock.Mock(return_value=self._make_github())
        result = config.get_latest_tag(beta=True)
        assert result == "beta/1.0-Beta_2"

    def test_get_latest_tag__beta_not_found(self):
        config = BaseProjectConfig(UniversalConfig())
        github = self._make_github()
        github.repositories["CumulusCI"]._releases = []
        config.get_github_api = mock.Mock(return_value=github)
        with pytest.raises(GithubException):
            config.get_latest_tag(beta=True)

    def test_get_latest_tag__repo_not_found(self):
        config = BaseProjectConfig(UniversalConfig())
        github = self._make_github()
        github.repositories["CumulusCI"] = None
        config.get_github_api = mock.Mock(return_value=github)
        with pytest.raises(GithubException):
            config.get_latest_tag()

    def test_get_latest_tag__release_not_found(self):
        config = BaseProjectConfig(UniversalConfig())
        github = self._make_github()
        github.repositories["CumulusCI"]._releases = []
        config.get_github_api = mock.Mock(return_value=github)
        with pytest.raises(GithubException):
            config.get_latest_tag()

    def test_get_latest_version(self):
        config = BaseProjectConfig(
            UniversalConfig(),
            {
                "project": {
                    "git": {"prefix_beta": "beta/", "prefix_release": "release/"}
                }
            },
        )
        config.get_github_api = mock.Mock(return_value=self._make_github())
        result = config.get_latest_version()
        assert result == "1.1"

    def test_get_latest_version_beta(self):
        config = BaseProjectConfig(
            UniversalConfig(),
            {
                "project": {
                    "git": {"prefix_beta": "beta/", "prefix_release": "release/"}
                }
            },
        )
        config.get_github_api = mock.Mock(return_value=self._make_github())
        result = config.get_latest_version(beta=True)
        assert result == "1.0 (Beta 2)"

    def test_get_previous_version(self):
        config = BaseProjectConfig(
            UniversalConfig(),
            {
                "project": {
                    "git": {"prefix_beta": "beta/", "prefix_release": "release/"}
                }
            },
        )
        config.get_github_api = mock.Mock(return_value=self._make_github())
        result = config.get_previous_version()
        assert result == "1.0"

    def test_config_project_path_no_repo_root(self):
        config = BaseProjectConfig(UniversalConfig())
        with temporary_dir():
            assert config.config_project_path is None

    def test_get_tag_for_version(self):
        config = BaseProjectConfig(
            UniversalConfig(), {"project": {"git": {"prefix_release": "release/"}}}
        )
        assert config.get_tag_for_version("beta/", "1.0") == "beta/1.0"

    def test_get_tag_for_version__1gp_beta(self):
        config = BaseProjectConfig(
            UniversalConfig(), {"project": {"git": {"prefix_beta": "beta/"}}}
        )
        assert config.get_tag_for_version("beta/", "1.0 (Beta 1)") == "beta/1.0-Beta_1"

    def test_get_tag_for_version__with_tag_prefix_option(self):
        config = BaseProjectConfig(UniversalConfig(), {})
        assert config.get_tag_for_version("custom/", "1.0") == "custom/1.0"

    def test_get_version_for_tag(self):
        config = BaseProjectConfig(
            UniversalConfig(),
            {
                "project": {
                    "git": {"prefix_beta": "beta/", "prefix_release": "release/"}
                }
            },
        )
        assert config.get_version_for_tag("release/1.0") == "1.0"

    def test_get_version_for_tag_invalid_beta(self):
        config = BaseProjectConfig(
            UniversalConfig(),
            {
                "project": {
                    "git": {"prefix_beta": "beta/", "prefix_release": "release/"}
                }
            },
        )
        assert config.get_version_for_tag("beta/invalid-format") is None

    def test_check_keychain(self):
        config = BaseProjectConfig(UniversalConfig())
        with pytest.raises(KeychainNotFound):
            config._check_keychain()

    def test_get_task__included_source(self):
        universal_config = UniversalConfig()
        with temporary_dir() as d:
            touch("cumulusci.yml")
            project_config = BaseProjectConfig(
                universal_config,
                {"sources": {"test": {"path": d}}},
                repo_info={"root": Path(__file__).parent.absolute()},
            )
            task_config = project_config.get_task("test:log")
        assert task_config.project_config is not project_config
        assert isinstance(task_config.project_config.source, LocalFolderSource)

    def test_get_flow__included_source(self):
        universal_config = UniversalConfig()
        with temporary_dir() as d:
            touch("cumulusci.yml")
            project_config = BaseProjectConfig(
                universal_config,
                {"sources": {"test": {"path": d}}},
                repo_info={"root": Path(__file__).parent.absolute()},
            )
            flow_config = project_config.get_flow("test:dev_org")
        assert flow_config.project_config is not project_config
        assert isinstance(flow_config.project_config.source, LocalFolderSource)

    def test_get_namespace__not_found(self):
        universal_config = UniversalConfig()
        project_config = BaseProjectConfig(universal_config)
        with pytest.raises(NamespaceNotFoundError):
            project_config.get_namespace("test")

    def test_include_source__bad_spec(self):
        universal_config = UniversalConfig()
        project_config = BaseProjectConfig(
            universal_config, {"sources": {"test": {"foo": "some_nonsense"}}}
        )
        with pytest.raises(ValueError, match="Invalid source spec"):
            project_config.include_source(project_config.get_namespace("test"))

    def test_include_source__cached(self):
        universal_config = UniversalConfig()
        project_config = BaseProjectConfig(
            universal_config,
            repo_info={"root": os.getcwd()},
        )
        with temporary_dir() as d:
            touch("cumulusci.yml")
            other1 = project_config.include_source(LocalFolderSourceModel(path=d))
            other2 = project_config.include_source(LocalFolderSourceModel(path=d))
        assert other1 is other2

    @mock.patch("cumulusci.core.config.project_config.GitHubSource")
    def test_include_source__github(self, source):
        source.return_value = expected_result = mock.Mock()
        expected_result.fetch.return_value.repo_root = "/whatever"
        universal_config = UniversalConfig()
        project_config = BaseProjectConfig(universal_config)
        other_config = project_config.include_source(
            GitHubSourceModel(github="foo/bar")
        )
        assert other_config.source is expected_result

    def test_relpath(self):
        universal_config = UniversalConfig()
        project_config = BaseProjectConfig(universal_config)
        assert project_config.relpath(os.path.abspath(".")) == "."

    def test_validate_package_api_version_valid(self):
        """We stringify the float 46.0 as this is what will occur when
        it is formatted into API URLS. This also negates the need to
        test an explicit string (i.e. if this passes we know that '46.0'
        will also pass)."""
        project_config = BaseProjectConfig(UniversalConfig())
        project_config.config["project"]["package"]["api_version"] = str(46.0)
        project_config._validate_package_api_format()

    def test_validate_package_api_version_invalid(self):
        project_config = BaseProjectConfig(UniversalConfig())
        project_config.config["project"]["package"]["api_version"] = str([1, 2, 3])
        with pytest.raises(ConfigError):
            project_config._validate_package_api_format()

        project_config.config["project"]["package"]["api_version"] = "9"
        with pytest.raises(ConfigError):
            project_config._validate_package_api_format()

        project_config.config["project"]["package"]["api_version"] = "9.0"
        with pytest.raises(ConfigError):
            project_config._validate_package_api_format()

        project_config.config["project"]["package"]["api_version"] = "45"
        with pytest.raises(ConfigError):
            project_config._validate_package_api_format()

        project_config.config["project"]["package"]["api_version"] = "45."
        with pytest.raises(ConfigError):
            project_config._validate_package_api_format()

        project_config.config["project"]["package"]["api_version"] = "45.00"
        with pytest.raises(ConfigError):
            project_config._validate_package_api_format()

    @mock.patch("cumulusci.core.config.project_config.git_path")
    def test_git_config_remote_origin_line(self, git_path):
        git_config_file = "test_git_config_file"
        git_path.return_value = git_config_file

        with open(git_config_file, "w") as f:
            f.writelines(
                [
                    '[branch "feature-1"]\n',
                    "\tremote = origin\n",
                    "\tmerge = refs/heads/feature-1\n",
                    '[remote "origin"]\n',
                    "\tfetch = +refs/heads/*:refs/remotes/origin/*\n",
                ]
            )

        project_config = BaseProjectConfig(UniversalConfig())
        actual_line = project_config.git_config_remote_origin_url()
        assert actual_line is None  # no url under [remote "origin"]

        with open(git_config_file, "a") as f:
            f.write("\turl = some.url.here\n")

        actual_line = project_config.git_config_remote_origin_url()
        assert actual_line == "some.url.here"

        os.remove(git_config_file)
        actual_line = project_config.git_config_remote_origin_url()
        assert actual_line is None  # no config file present

    def test_default_package_path(self):
        config = BaseProjectConfig(UniversalConfig())
        assert str(config.default_package_path.relative_to(config.repo_root)) == "src"

    def test_default_package_path__sfdx(self):
        with temporary_dir() as path:
            pathlib.Path(path, ".git").mkdir()
            with pathlib.Path(path, "cumulusci.yml").open("w") as f:
                yaml.dump({"project": {"source_format": "sfdx"}}, f)
            with pathlib.Path(path, "sfdx-project.json").open("w") as f:
                json.dump(
                    {"packageDirectories": [{"path": "force-app", "default": True}]}, f
                )
            config = BaseProjectConfig(UniversalConfig())
            assert (
                str(config.default_package_path.relative_to(config.repo_root))
                == "force-app"
            )


class TestBaseTaskFlowConfig:
    def setup_method(self):
        self.task_flow_config = BaseTaskFlowConfig(
            {
                "tasks": {
                    "deploy": {"description": "Deploy Task", "class_path": "my.path"},
                    "no_path": {"description": "Only a Description"},
                    "manage": {},
                    "control": {},
                },
                "flows": {
                    "coffee": {"description": "Coffee Flow"},
                    "juice": {"description": "Juice Flow"},
                },
            }
        )

    def test_list_tasks(self):
        tasks = self.task_flow_config.list_tasks()
        assert len(tasks) == 4
        deploy = [task for task in tasks if task["name"] == "deploy"][0]
        assert deploy["description"] == "Deploy Task"

    def test_get_task(self):
        task = self.task_flow_config.get_task("deploy")
        assert isinstance(task, BaseConfig)
        assert ("description", "Deploy Task") in task.config.items()

    def test_get_task__no_class_path(self):
        with pytest.raises(
            CumulusCIException, match="Task has no class_path defined: no_path"
        ):
            self.task_flow_config.get_task("no_path")

    def test_get_task__no_config_found(self):
        with pytest.raises(
            CumulusCIException, match="No configuration found for task: manage"
        ):
            self.task_flow_config.get_task("manage")

    def test_no_task__no_suggestion(self):
        with pytest.raises(
            TaskNotFoundError, match="Task not found: robotic_superstar"
        ):
            self.task_flow_config.get_task("robotic_superstar")

    def test_no_task__with_suggestion(self):
        with pytest.raises(
            TaskNotFoundError, match='Task not found: mange. Did you mean "manage"?'
        ):
            self.task_flow_config.get_task("mange")

    def test_get_flow(self):
        flow = self.task_flow_config.get_flow("coffee")
        assert isinstance(flow, BaseConfig)
        assert ("description", "Coffee Flow") in flow.config.items()

    def test_no_flow(self):
        with pytest.raises(FlowNotFoundError):
            self.task_flow_config.get_flow("water")

    def test_list_flows(self):
        flows = self.task_flow_config.list_flows()
        assert len(flows) == 2
        coffee = [flow for flow in flows if flow["name"] == "coffee"][0]
        assert coffee["description"] == "Coffee Flow"

    def test_suggested_name(self):
        flows = self.task_flow_config.flows
        assert len(flows) == 2
        error_msg = self.task_flow_config.get_suggested_name("bofee", flows)
        assert "coffee" in error_msg


class TestOrgConfig:
    @mock.patch("cumulusci.core.config.org_config.OrgConfig.OAuth2Client")
    def test_refresh_oauth_token(self, OAuth2Client):
        config = OrgConfig(
            {
                "refresh_token": mock.sentinel.refresh_token,
                "instance_url": "http://instance_url_111.com",
            },
            "test",
        )
        project_config = BaseProjectConfig(UniversalConfig())
        keychain = BaseProjectKeychain(project_config, None)
        config._load_userinfo = mock.Mock()
        config._load_orginfo = mock.Mock()

        refresh_token = mock.Mock(return_value={"access_token": "asdf"})
        OAuth2Client.return_value = mock.Mock(refresh_token=refresh_token)

        config.refresh_oauth_token(keychain)

        client_config = OAuth2Client.call_args[0][0]
        assert client_config.client_id == DEFAULT_CONNECTED_APP.client_id
        refresh_token.assert_called_once_with(mock.sentinel.refresh_token)

    @mock.patch("cumulusci.core.config.org_config.OrgConfig.OAuth2Client")
    def test_refresh_oauth_token__other_connected_app(self, OAuth2Client):
        config = OrgConfig(
            {
                "connected_app": "other",
                "refresh_token": mock.sentinel.refresh_token,
                "instance_url": "http://instance_url_111.com",
            },
            "test",
        )
        project_config = BaseProjectConfig(UniversalConfig())
        keychain = BaseProjectKeychain(project_config, None)
        keychain.set_service(
            "connected_app",
            "other",
            ServiceConfig(
                {
                    "login_url": "https://other",
                    "callback_url": "http://localhost:8080/callback",
                    "client_id": "OTHER_ID",
                    "client_secret": "OTHER_SECRET",
                }
            ),
        )
        config._load_userinfo = mock.Mock()
        config._load_orginfo = mock.Mock()

        refresh_token = mock.Mock(return_value={"access_token": "asdf"})
        OAuth2Client.return_value = mock.Mock(refresh_token=refresh_token)

        config.refresh_oauth_token(keychain)

        client_config = OAuth2Client.call_args[0][0]
        assert client_config.client_id == "OTHER_ID"
        refresh_token.assert_called_once_with(mock.sentinel.refresh_token)

    @responses.activate
    def test_load_user_info__bad_json(self):
        config = OrgConfig(
            {
                "refresh_token": mock.sentinel.refresh_token,
                "instance_url": "http://instance_url_111.com",
            },
            "test",
        )
        keychain = mock.Mock()
        keychain.get_service.return_value = mock.Mock(
            client_id="asdf", client_secret="asdf"
        )

        responses.add(
            responses.POST, "http://instance_url_111.com/services/oauth2/token"
        )
        with pytest.raises(CumulusCIException) as e:
            config.refresh_oauth_token(keychain)
        assert "Cannot decode" in str(e.value)

    def test_refresh_oauth_token_no_connected_app(self):
        config = OrgConfig({}, "test")
        with pytest.raises(AttributeError):
            config.refresh_oauth_token(None)

    def test_refresh_oauth_token__bad_connected_app(self):
        org_config = OrgConfig({"connected_app": "bogus"}, "test")
        project_config = BaseProjectConfig(UniversalConfig())
        keychain = BaseProjectKeychain(project_config, None)
        with pytest.raises(ServiceNotConfigured, match="no longer configured"):
            org_config.refresh_oauth_token(keychain)

    @mock.patch("jwt.encode", mock.Mock(return_value="JWT"))
    @responses.activate
    def test_refresh_oauth_token__jwt(self):
        responses.add(
            "POST",
            "https://login.salesforce.com/services/oauth2/token",
            json={
                "access_token": "TOKEN",
                "instance_url": "https://na00.salesforce.com",
            },
        )
        with mock.patch.dict(
            os.environ,
            {"SFDX_CLIENT_ID": "some client id", "SFDX_HUB_KEY": "some private key"},
        ):
            config = OrgConfig({}, "test")
            config._load_userinfo = mock.Mock()
            config._load_orginfo = mock.Mock()
            config.refresh_oauth_token(None)
            assert config.access_token == "TOKEN"

    @mock.patch("jwt.encode", mock.Mock(return_value="JWT"))
    @responses.activate
    def test_refresh_oauth_token__jwt_sandbox(self):
        responses.add(
            "POST",
            "https://cs00.salesforce.com/services/oauth2/token",
            json={
                "access_token": "TOKEN",
                "instance_url": "https://cs00.salesforce.com",
            },
        )
        with mock.patch.dict(
            os.environ,
            {"SFDX_CLIENT_ID": "some client id", "SFDX_HUB_KEY": "some private key"},
        ):
            config = OrgConfig(
                {
                    "instance_url": "https://cs00.salesforce.com",
                },
                "test",
            )
            config._load_userinfo = mock.Mock()
            config._load_orginfo = mock.Mock()
            config.refresh_oauth_token(None)
            assert config.access_token == "TOKEN"

    @mock.patch("jwt.encode", mock.Mock(return_value="JWT"))
    @responses.activate
    def test_refresh_oauth_token__jwt_sandbox_instanceless_url(self):
        responses.add(
            "POST",
            "https://nonobvious--sandbox.my.salesforce.com/services/oauth2/token",
            json={
                "access_token": "TOKEN",
                "instance_url": "https://nonobvious--sandbox.my.salesforce.com",
            },
        )
        with mock.patch.dict(
            os.environ,
            {"SFDX_CLIENT_ID": "some client id", "SFDX_HUB_KEY": "some private key"},
        ):
            config = OrgConfig(
                {
                    "instance_url": "https://nonobvious--sandbox.my.salesforce.com",
                    "id": "https://test.salesforce.com/asdf",
                },
                "test",
            )
            config._load_userinfo = mock.Mock()
            config._load_orginfo = mock.Mock()
            config.refresh_oauth_token(None)
            assert config.access_token == "TOKEN"

    def test_lightning_base_url__instance(self):
        config = OrgConfig({"instance_url": "https://na01.salesforce.com"}, "test")
        assert config.lightning_base_url == "https://na01.lightning.force.com"

    def test_lightning_base_url__scratch_org(self):
        config = OrgConfig(
            {"instance_url": "https://foo.cs42.my.salesforce.com"}, "test"
        )
        assert config.lightning_base_url == "https://foo.lightning.force.com"

    def test_lightning_base_url__mydomain(self):
        config = OrgConfig({"instance_url": "https://foo.my.salesforce.com"}, "test")
        assert config.lightning_base_url == "https://foo.lightning.force.com"

    @responses.activate
    def test_get_salesforce_version(self):
        responses.add(
            "GET",
            "https://na01.salesforce.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )
        config = OrgConfig({"instance_url": "https://na01.salesforce.com"}, "test")
        config.access_token = "TOKEN"
        assert config.latest_api_version == CURRENT_SF_API_VERSION

    @responses.activate
    def test_get_salesforce_version_bad_json(self):
        responses.add("GET", "https://na01.salesforce.com/services/data", "NOTJSON!")
        config = OrgConfig({"instance_url": "https://na01.salesforce.com"}, "test")
        config.access_token = "TOKEN"
        with pytest.raises(CumulusCIException) as e:
            assert config.latest_api_version == "42.0"
        assert "NOTJSON" in str(e.value)

    @responses.activate
    def test_get_salesforce_version_weird_json(self):
        responses.add(
            "GET", "https://na01.salesforce.com/services/data", json=["NOTADICT"]
        )
        config = OrgConfig({"instance_url": "https://na01.salesforce.com"}, "test")
        config.access_token = "TOKEN"
        with pytest.raises(CumulusCIException) as e:
            assert config.latest_api_version == "42.0"
        assert "NOTADICT" in str(e.value)

    def test_start_url(self):
        config = OrgConfig(
            {"instance_url": "https://na01.salesforce.com", "access_token": "TOKEN"},
            "test",
        )
        assert (
            "https://na01.salesforce.com/secur/frontdoor.jsp?sid=TOKEN"
            == config.start_url
        )

    def test_user_id(self):
        config = OrgConfig({"id": "org/user"}, "test")
        assert config.user_id == "user"

    def test_can_delete(self):
        config = OrgConfig({}, "test")
        assert not config.can_delete()

    @responses.activate
    def test_load_orginfo(self):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )
        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/Organization/OODxxxxxxxxxxxx",
            json={
                "OrganizationType": "Enterprise Edition",
                "IsSandbox": False,
                "InstanceName": "cs420",
                "NamespacePrefix": "ns",
            },
        )

        config._load_orginfo()

        assert config.org_type == "Enterprise Edition"
        assert config.is_sandbox is False
        assert config.organization_sobject is not None
        assert config.namespace == "ns"

    @responses.activate
    def test_get_community_info__cached(self):
        """Verify that get_community_info returns data from the cache"""
        config = OrgConfig({}, "test")
        config._community_info_cache = {"Kōkua": {"name": "Kōkua"}}
        info = config.get_community_info("Kōkua")
        assert info["name"] == "Kōkua"

    @responses.activate
    def test_get_community_info__fetch_if_not_in_cache(self):
        """Verify that the internal cache is automatically refreshed

        The cache should be refreshed automatically if the requested community
        is not in the cache.
        """
        responses.add(
            "GET",
            "https://test/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        responses.add(
            "GET",
            f"https://test/services/data/v{CURRENT_SF_API_VERSION}/connect/communities",
            json={"communities": [{"name": "Kōkua"}]},
        )

        config = OrgConfig(
            {"instance_url": "https://test", "access_token": "TOKEN"}, "test"
        )
        config._community_info_cache = {}
        info = config.get_community_info("Kōkua")
        assert info["name"] == "Kōkua"

    @mock.patch("cumulusci.core.config.org_config.OrgConfig._fetch_community_info")
    def test_community_info_force_refresh(self, mock_fetch):
        """Verify that the force_refresh parameter has an effect"""
        mock_fetch.return_value = {"Kōkua": {"name": "Kōkua"}}
        config = OrgConfig({}, "test")

        # With the cache seeded with the target community, first
        # verify that the cache isn't refreshed automatically
        config._community_info_cache = {"Kōkua": {"name": "Kōkua"}}
        config.get_community_info("Kōkua")
        mock_fetch.assert_not_called()

        # Now, set force_refresh and make sure it is refreshed
        config.get_community_info("Kōkua", force_refresh=True)
        mock_fetch.assert_called()

    @mock.patch("cumulusci.core.config.org_config.OrgConfig._fetch_community_info")
    def test_community_info_exception(self, mock_fetch):
        """Verify an exception is thrown when the community doesn't exist"""
        config = OrgConfig({}, "test")
        expected_exception = "Unable to find community information for 'bogus'"
        with pytest.raises(Exception, match=expected_exception):
            config.get_community_info("bogus")

    MOCK_TOOLING_PACKAGE_RESULTS = [
        {
            "size": 2,
            "totalSize": 2,
            "done": True,
            "records": [
                {
                    "SubscriberPackage": {
                        "Id": "03350000000DEz4AAG",
                        "NamespacePrefix": "GW_Volunteers",
                    },
                    "SubscriberPackageVersionId": "04t1T00000070yqQAA",
                },
                {
                    "SubscriberPackage": {
                        "Id": "03350000000DEz5AAG",
                        "NamespacePrefix": "GW_Volunteers",
                    },
                    "SubscriberPackageVersionId": "04t000000000001AAA",
                },
                {
                    "SubscriberPackage": {
                        "Id": "03350000000DEz7AAG",
                        "NamespacePrefix": "TESTY",
                    },
                    "SubscriberPackageVersionId": "04t000000000002AAA",
                },
                {
                    "SubscriberPackage": {
                        "Id": "03350000000DEz4AAG",
                        "NamespacePrefix": "blah",
                    },
                    "SubscriberPackageVersionId": "04t0000000BOGUSAAA",
                },
                {
                    "SubscriberPackage": {
                        "Id": "03350000000DEz8AAG",
                        "NamespacePrefix": "error",
                    },
                    "SubscriberPackageVersionId": "04t0000000ERRORAAA",
                },
            ],
        },
        {
            "size": 1,
            "totalSize": 1,
            "done": True,
            "records": [
                {
                    "Id": "04t1T00000070yqQAA",
                    "MajorVersion": 3,
                    "MinorVersion": 119,
                    "PatchVersion": 0,
                    "BuildNumber": 5,
                    "IsBeta": False,
                }
            ],
        },
        {
            "size": 1,
            "totalSize": 1,
            "done": True,
            "records": [
                {
                    "Id": "04t000000000001AAA",
                    "MajorVersion": 12,
                    "MinorVersion": 0,
                    "PatchVersion": 1,
                    "BuildNumber": 1,
                    "IsBeta": False,
                }
            ],
        },
        {
            "size": 1,
            "totalSize": 1,
            "done": True,
            "records": [
                {
                    "Id": "04t000000000002AAA",
                    "MajorVersion": 1,
                    "MinorVersion": 10,
                    "PatchVersion": 0,
                    "BuildNumber": 5,
                    "IsBeta": True,
                }
            ],
        },
        {"size": 0, "totalSize": 0, "done": True, "records": []},
        SalesforceError(None, None, None, None),
    ]

    @mock.patch("cumulusci.core.config.org_config.OrgConfig.salesforce_client")
    def test_installed_packages(self, sf):
        config = OrgConfig({}, "test")
        sf.restful.side_effect = self.MOCK_TOOLING_PACKAGE_RESULTS

        expected = {
            "GW_Volunteers": [
                VersionInfo("04t1T00000070yqQAA", StrictVersion("3.119")),
                VersionInfo("04t000000000001AAA", StrictVersion("12.0.1")),
            ],
            "GW_Volunteers@3.119": [
                VersionInfo("04t1T00000070yqQAA", StrictVersion("3.119"))
            ],
            "GW_Volunteers@12.0.1": [
                VersionInfo("04t000000000001AAA", StrictVersion("12.0.1"))
            ],
            "TESTY": [VersionInfo("04t000000000002AAA", StrictVersion("1.10.0b5"))],
            "TESTY@1.10b5": [
                VersionInfo("04t000000000002AAA", StrictVersion("1.10.0b5"))
            ],
            "03350000000DEz4AAG": [
                VersionInfo("04t1T00000070yqQAA", StrictVersion("3.119"))
            ],
            "03350000000DEz5AAG": [
                VersionInfo("04t000000000001AAA", StrictVersion("12.0.1"))
            ],
            "03350000000DEz7AAG": [
                VersionInfo("04t000000000002AAA", StrictVersion("1.10.0b5"))
            ],
        }
        # get it twice so we can make sure it is cached
        assert config.installed_packages == expected
        assert config.installed_packages == expected
        sf.restful.assert_called()

        sf.restful.reset_mock()
        sf.restful.side_effect = self.MOCK_TOOLING_PACKAGE_RESULTS
        config.reset_installed_packages()
        assert config.installed_packages == expected
        sf.restful.assert_called()

    @mock.patch("cumulusci.core.config.org_config.OrgConfig.salesforce_client")
    def test_has_minimum_package_version(self, sf):
        config = OrgConfig({}, "test")
        sf.restful.side_effect = self.MOCK_TOOLING_PACKAGE_RESULTS

        assert config.has_minimum_package_version("TESTY", "1.9")
        assert config.has_minimum_package_version("TESTY", "1.10b5")
        assert not config.has_minimum_package_version("TESTY", "1.10b6")
        assert not config.has_minimum_package_version("TESTY", "1.10")
        assert not config.has_minimum_package_version("npsp", "1.0")

        assert config.has_minimum_package_version("03350000000DEz4AAG", "3.119")

        with pytest.raises(CumulusCIException):
            config.has_minimum_package_version("GW_Volunteers", "1.0")

    def test_orginfo_cache_dir_global(self):
        config = OrgConfig(
            {
                "instance_url": "http://zombo.com/welcome",
                "username": "test-example@example.com",
            },
            "test",
            keychain=DummyKeychain(),
            global_org=True,
        )
        with TemporaryDirectory() as t:
            with mock.patch(
                "cumulusci.tests.util.DummyKeychain.global_config_dir", Path(t)
            ):
                with config.get_orginfo_cache_dir("foo") as directory:
                    assert directory.exists()
                    assert str(t) in directory, (t, directory)
                    assert (
                        str(directory)
                        .replace("\\", "/")
                        .endswith("orginfo/zombo.com__test-example__example.com/foo")
                    ), str(directory).replace("\\", "/")
                    foo = directory / "Foo.txt"
                    with foo.open("w") as f:
                        f.write("Bar")
                    with foo.open("r") as f:
                        assert f.read() == "Bar"

    def test_orginfo_cache_dir_local(self):
        config = OrgConfig(
            {
                "instance_url": "http://zombo.com/welcome",
                "username": "test-example@example.com",
            },
            "test",
            keychain=DummyKeychain(),
            global_org=False,
        )
        with TemporaryDirectory() as t:
            with mock.patch("cumulusci.tests.util.DummyKeychain.cache_dir", Path(t)):

                with config.get_orginfo_cache_dir("bar") as directory:
                    assert str(t) in directory, (t, directory)
                    assert (
                        str(directory)
                        .replace("\\", "/")
                        .endswith("orginfo/zombo.com__test-example__example.com/bar")
                    )
                    assert directory.exists()
                    foo = directory / "Foo.txt"
                    with foo.open("w") as f:
                        f.write("Bar")
                    with foo.open("r") as f:
                        assert f.read() == "Bar"

    @responses.activate
    def test_is_person_accounts_enabled__not_enabled(self):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )
        assert (
            config._is_person_accounts_enabled is None
        ), "_is_person_accounts_enabled should be initialized as None"

        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/Account/describe",
            json={"fields": [{"name": "Id"}]},
        )

        # Verify checks describe if _is_person_accounts_enabled is None.
        actual = config.is_person_accounts_enabled

        assert actual is False, ""
        assert actual == config._is_person_accounts_enabled

        # Verify subsequent calls return cached value.
        config._is_person_accounts_enabled = True

        assert config._is_person_accounts_enabled == config.is_person_accounts_enabled

    @responses.activate
    def test_is_person_accounts_enabled__is_enabled(self):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )
        assert (
            config._is_person_accounts_enabled is None
        ), "_is_person_accounts_enabled should be initialized as None"

        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/Account/describe",
            json={"fields": [{"name": "Id"}, {"name": "IsPersonAccount"}]},
        )

        # Verify checks describe if _is_person_accounts_enabled is None.
        actual = config.is_person_accounts_enabled

        assert actual is True, ""
        assert actual == config._is_person_accounts_enabled

        # Verify subsequent calls return cached value.
        config._is_person_accounts_enabled = False

        assert config._is_person_accounts_enabled == config.is_person_accounts_enabled

    @responses.activate
    def test_is_multi_currency_enabled__not_enabled(self):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )
        assert (
            config._multiple_currencies_is_enabled is False
        ), "_multiple_currencies_is_enabled should be initialized as False"

        # Login call.
        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        # CurrencyType describe() call.
        # Since Multiple Currencies is not enabled, CurrencyType Sobject is not exposed.
        # Therefore, the describe call will result in a 404.
        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/CurrencyType/describe",
            status=404,
            json={
                "errorCode": "NOT_FOUND",
                "message": "The requested resource does not exist",
            },
        )

        # Add a second 404 to demonstrate we always check the describe until we detect Multiple Currencies is enabled.  From then on, we cache the fact that Multiple Currencies is enabled knowing Multiple Currencies cannot be disabled.
        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/CurrencyType/describe",
            status=404,
            json={
                "errorCode": "NOT_FOUND",
                "message": "The requested resource does not exist",
            },
        )

        # Check 1: is_multiple_currencies_enabled should be False since the CurrencyType describe gives a 404.
        actual = config.is_multiple_currencies_enabled
        assert (
            actual is False
        ), "config.is_multiple_currencies_enabled should be False since the CurrencyType describe returns a 404."
        assert (
            config._multiple_currencies_is_enabled is False
        ), "config._multiple_currencies_is_enabled should still be False since the CurrencyType describe returns a 404."

        # Check 2: We should still get the CurrencyType describe since we never cached that multiple currencies is enabled.
        actual = config.is_multiple_currencies_enabled
        assert (
            actual is False
        ), "config.is_multiple_currencies_enabled should be False since the CurrencyType describe returns a 404."
        assert (
            config._multiple_currencies_is_enabled is False
        ), "config._multiple_currencies_is_enabled should still be False since the CurrencyType describe returns a 404."

        # We should have made 3 calls: 1 token call + 2 describe calls
        assert len(responses.calls) == 1 + 2

    @responses.activate
    def test_is_multi_currency_enabled__is_enabled(self):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )

        assert (
            config._multiple_currencies_is_enabled is False
        ), "_multiple_currencies_is_enabled should be initialized as False"

        # Token call.
        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        # CurrencyType describe() call.
        # Since Multiple Currencies is enabled, so the describe call returns a 200.
        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/CurrencyType/describe",
            json={
                # The actual payload doesn't matter; only matters is we get a 200.
            },
        )

        # Check 1: is_multiple_currencies_enabled should be True since the CurrencyType describe gives a 200.
        actual = config.is_multiple_currencies_enabled
        assert (
            actual is True
        ), "config.is_multiple_currencies_enabled should be True since the CurrencyType describe returns a 200."
        assert (
            config._multiple_currencies_is_enabled is True
        ), "config._multiple_currencies_is_enabled should be True since the CurrencyType describe returns a 200."

        # Check 2: We should have cached that Multiple Currencies is enabled, so we should not make a 2nd descrobe call. This is ok to cache since Multiple Currencies cannot be disabled.
        actual = config.is_multiple_currencies_enabled
        assert (
            actual is True
        ), "config.is_multiple_currencies_enabled should be True since the our cached value in _multiple_currencies_is_enabled is True."
        assert (
            config._multiple_currencies_is_enabled is True
        ), "config._multiple_currencies_is_enabled should still be True."

        # We should have made 2 calls: 1 token call + 1 describe call
        assert len(responses.calls) == 1 + 1

    @responses.activate
    def test_is_advanced_currency_management_enabled__multiple_currencies_not_enabled(
        self,
    ):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )

        # Token call.
        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        # DatedConversionRate describe() call.
        # Since Multiple Currencies is not enabled, DatedConversionRate Sobject is not exposed.
        # Therefore, the describe call will result in a 404.
        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/DatedConversionRate/describe",
            status=404,
            json={
                "errorCode": "NOT_FOUND",
                "message": "The requested resource does not exist",
            },
        )

        # is_advanced_currency_management_enabled should be False since:
        # - DatedConversionRate describe gives a 404 implying the Sobject is not exposed becuase Multiple Currencies is not enabled.
        actual = config.is_advanced_currency_management_enabled
        assert (
            actual is False
        ), "config.is_advanced_currency_management_enabled should be False since the describe gives a 404."

        # We should have made 2 calls: 1 token call + 1 describe call
        assert len(responses.calls) == 1 + 1

    @responses.activate
    def test_is_advanced_currency_management_enabled__multiple_currencies_enabled__acm_not_enabled(
        self,
    ):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )

        # Token call.
        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        # DatedConversionRate describe() call.
        # Since Multiple Currencies is enabled, so the describe call returns a 200.
        # However, ACM is not enabled so DatedConversionRate is not createable.
        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/DatedConversionRate/describe",
            json={"createable": False},
        )

        # is_advanced_currency_management_enabled should be False:
        # - DatedConversionRate describe gives a 200, so the Sobject is exposed (because Multiple Currencies is enabled).
        # - But DatedConversionRate is not creatable implying ACM is not enabled.
        actual = config.is_advanced_currency_management_enabled
        assert (
            actual is False
        ), 'config.is_advanced_currency_management_enabled should be False since though the describe gives a 200, the describe is not "createable".'

        # We should have made 2 calls: 1 token call + 1 describe call
        assert len(responses.calls) == 1 + 1

    @responses.activate
    def test_is_advanced_currency_management_enabled__multiple_currencies_enabled__acm_enabled(
        self,
    ):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )

        # Token call.
        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        # DatedConversionRate describe() call.
        # Since Multiple Currencies is enabled, so the describe call returns a 200.
        # However, ACM is not enabled so DatedConversionRate is not createable.
        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/DatedConversionRate/describe",
            json={"createable": True},
        )

        # is_advanced_currency_management_enabled should be False:
        # - DatedConversionRate describe gives a 200, so the Sobject is exposed (because Multiple Currencies is enabled).
        # - But DatedConversionRate is not creatable implying ACM is not enabled.
        actual = config.is_advanced_currency_management_enabled
        assert (
            actual is True
        ), 'config.is_advanced_currency_management_enabled should be False since both the describe gives a 200 and the describe is "createable".'

        # We should have made 2 calls: 1 token call + 1 describe call
        assert len(responses.calls) == 1 + 1

    @responses.activate
    def test_is_survey_advanced_features_enabled(self):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )

        # Token call.
        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        # describe()
        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/PermissionSet/describe",
            json={"fields": [{"name": "PermissionsAllowSurveyAdvancedFeatures"}]},
        )

        assert config.is_survey_advanced_features_enabled

    @responses.activate
    def test_is_survey_advanced_features_enabled__not_enabled(self):
        config = OrgConfig(
            {
                "instance_url": "https://example.com",
                "access_token": "TOKEN",
                "id": "OODxxxxxxxxxxxx/user",
            },
            "test",
        )

        # Token call.
        responses.add(
            "GET",
            "https://example.com/services/data",
            json=[{"version": CURRENT_SF_API_VERSION}],
        )

        # describe()
        responses.add(
            "GET",
            f"https://example.com/services/data/v{CURRENT_SF_API_VERSION}/sobjects/PermissionSet/describe",
            json={"fields": [{"name": "foo"}]},
        )

        assert not config.is_survey_advanced_features_enabled

    def test_resolve_04t_dependencies(self):
        config = OrgConfig({}, "test")
        config._installed_packages = {
            "dep@1.0": [VersionInfo("04t000000000001AAA", "1.0")]
        }
        result = config.resolve_04t_dependencies(
            [PackageNamespaceVersionDependency(namespace="dep", version="1.0")]
        )
        assert result == [PackageVersionIdDependency(version_id="04t000000000001AAA")]

    def test_resolve_04t_dependencies__not_installed(self):
        config = OrgConfig({}, "test")
        config._installed_packages = {}
        with pytest.raises(DependencyResolutionError):
            config.resolve_04t_dependencies(
                [PackageNamespaceVersionDependency(namespace="dep", version="1.0")]
            )



# -*- coding: utf-8 -*-
"""Format and display CLI output.

Classes:
    CliTable: Pretty prints tabular data to stdout, via Rich's Console API
"""
import os
from typing import Any, List, Union

import rich
from rich import box, print
from rich.console import Console
from rich.style import Style
from rich.table import Column, Table

CHECKMARK = "[green]:heavy_check_mark:" if os.name == "posix" else "+"
CROSSMARK = "[red]:cross_mark:" if os.name == "posix" else "-"


class CliTable:
    """Format and print data to the command line in tabular form.
    Attributes:

    * PICTOGRAM_TRUE: True boolean values are replaced with this string.
    * PICTOGRAM_FALSE = False boolean values are replaced with this string.

    Methods:
    * echo: Print the table data to stdout using rich.Console.print
    Class Methods:
    * columnify_headers: Convert a list of strs to Columns
    """

    PICTOGRAM_TRUE = CHECKMARK
    PICTOGRAM_FALSE = " "

    def __init__(
        self,
        data: List[List[Any]],
        title: str = None,
        dim_rows: List[int] = None,
        **kwargs,
    ):
        """Constructor.
        Args:
            data: Required. List[List] of data to format, with the heading as 0-th member.
            title: String to use as the table's title.
            dim_rows: List[int] of row indices to dim.
        """

        headers: List[Column] = self.columnify_headers(data[0])
        rows = [self._stringify_row(r) for r in data[1:]]
        self._table: Table = Table(*headers, title=title, box=box.SIMPLE, **kwargs)

        for idx, row in enumerate(rows):
            dim_row = idx + 1 in dim_rows if dim_rows else False
            self._table.add_row(*row, style=Style(dim=dim_row))

    def _stringify_row(self, row: list) -> List[str]:
        return [self._stringify_cell(cell) for cell in row]

    def _stringify_cell(self, cell: Any) -> str:
        if isinstance(cell, bool):
            cell = self.PICTOGRAM_TRUE if cell else self.PICTOGRAM_FALSE
        elif cell is None:
            cell = ""
        else:
            cell = str(cell)
        return cell

    def __rich__(self) -> str:
        return self._table

    def columnify_headers(self, headers: List[Union[str, Column]]) -> List[Column]:
        """
        Given a List[str, Column], return a list of Columns.

        Strings are instantiated with default styles, Columns are unchanged.
        """
        cols: List[Column] = []

        for header in headers:
            if isinstance(header, Column):
                cols.append(header)
            else:
                header: str = self._stringify_cell(header)
                cols.append(Column(header=header, overflow="fold"))
        return cols

    def echo(self, plain=False, box_style: box.Box = None):
        """
        Print this table to the global Console using console.print().
        """
        orig_box = self._table.box
        if plain:
            self._table.box = box.ASCII2
        else:
            self._table.box = box_style or orig_box
        console = rich.get_console()
        console.print(self._table)
        self._table.box = orig_box

    @property
    def table(self):
        return self._table

    def __str__(self):
        from io import StringIO

        console = Console(file=StringIO())
        tab_width = console.size.width - 21
        console.print(self._table, width=tab_width)
        return console.file.getvalue()


def _soql_table(results, truncated):
    if results:
        if truncated:
            assert results[-1] == truncated
            first_row = results[0]
            fake_row = {k: "" for k, v in first_row.items()}
            first_column = list(first_row)[0]
            fake_row[first_column] = truncated

            results[-1] = fake_row

        headings = list(results[0].keys())
        return CliTable(
            [headings] + [list(map(str, r.values())) for r in results],
        ).echo()
    else:
        return CliTable([["No results"]]).echo()


def _summarize(field):
    if field["referenceTo"]:
        allowed = field["referenceTo"]
    elif field["picklistValues"]:
        allowed = [value["value"] for value in field["picklistValues"]]
    else:
        allowed = field["type"]
    return (field["name"], allowed)


class SimpleSalesforceUIHelpers:
    def __init__(self, sf):
        self._sf = sf

    def query(self, query, format="table", include_deleted=False, max_rows=100):
        """Return the result of a Salesforce SOQL query.

        Arguments:

            * query -- the SOQL query to send to Salesforce, e.g.
                    SELECT Id FROM Lead WHERE Email = "waldo@somewhere.com"
            * include_deleted -- True if deleted records should be included
            * format -- one of these values:
                - "table" -- printable ASCII table (the default)
                - "obj" -- ordinary Python objects
                - "pprint" -- string in easily readable Python dict shape
                - "json" -- JSON
            * max_rows -- maximum rows to output, defaults to 100

        For example:
            query("select Name, Id from Account", format="pprint")
            contact_ids = query("select count(Id) from Contact", format="obj")
        """
        results = self._sf.query_all(query, include_deleted=include_deleted)["records"]

        if len(results) > max_rows:
            truncated = f"... truncated {len(results) - max_rows} rows"
            results = results[0:max_rows]
        else:
            truncated = False

        for result in results:
            if result.get("attributes"):
                del result["attributes"]

        if truncated:
            results.append(truncated)

        if format == "table":
            help_message = "Type help(query) to learn about other return formats or assigning the result."
            print(_soql_table(results, truncated))
            print()
            print(help_message)
            rc = None
        elif format == "obj":
            rc = results
        elif format == "pprint":
            from rich.pretty import pprint

            pprint(results)
            rc = None
        elif format == "json":
            from json import dumps

            rc = dumps(results, indent=2)
        else:
            raise TypeError(f"Unknown format `{format}`")

        return rc

    def describe(self, sobj_name, detailed=False, format="pprint"):
        """Describe an sobject.

        Arguments:

            sobj_name - sobject name to describe. e.g. "Account", "Contact"
            detailed - set to `True` to get detailed information about object
            format -- one of these values:
            - "pprint" -- string in easily readable Python dict shape (default)
            - "obj" -- ordinary Python objects

        For example:

            >>> describe("Account")
            >>> data = describe("Account", detailed=True, format=obj)
        """
        from pprint import pprint

        data = getattr(self._sf, sobj_name).describe()

        if detailed:
            rc = data
        else:
            rc = dict(_summarize(field) for field in data["fields"])

        if format == "pprint":
            pprint(rc)
        elif format == "obj":
            return rc
        else:
            raise TypeError(f"Unknown format {format}")



import random

from cumulusci.tasks.bulkdata.extract_dataset_utils.calculate_dependencies import (
    SObjDependency,
)
from cumulusci.tasks.bulkdata.generate_mapping_utils.dependency_map import DependencyMap
from cumulusci.tasks.bulkdata.generate_mapping_utils.mapping_transforms import (
    merge_matching_steps,
    recategorize_lookups,
    rename_record_type_fields,
    sort_steps,
)
from cumulusci.tasks.bulkdata.mapping_parser import MappingStep


class MappingTransformTesterBase:
    @classmethod
    def setup_class(_):
        mttb = MappingTransformTesterBase
        mttb.std_intertable_dependencies = [
            SObjDependency("Contact", "Account", "AccountId"),
            SObjDependency("Account", "Account", "ParentId"),
            SObjDependency("Opportunity", "Contact", "ContactId"),
            SObjDependency("Opportunity", "Account", "AccountId"),
            SObjDependency("Account", "Custom__c", "Custom__c"),
            SObjDependency("Custom__c", "Opportunity", "Opportunity__c"),
            SObjDependency("Custom_2__c", "Custom__c", "Custom__c"),
            SObjDependency("Custom_2__c", "Custom_3__c", "Custom_3__c"),
            SObjDependency("Custom_3__c", "Custom_2__c", "Custom_2__c"),
        ]
        mttb.std_table_names = [
            dep.table_name_from for dep in mttb.std_intertable_dependencies
        ] + ["Unrelated__c"]

        mttb.std_dep_map = DependencyMap(
            mttb.std_table_names, mttb.std_intertable_dependencies
        )


class TestSortSteps(MappingTransformTesterBase):
    def test_sort_steps(self):
        steps = [
            MappingStep(sf_object="Opportunity"),
            MappingStep(sf_object="Custom__c"),
            MappingStep(sf_object="Custom_2__c"),
            MappingStep(sf_object="Custom_3__c"),
            MappingStep(sf_object="Contact"),
            MappingStep(sf_object="Account"),
        ]

        for i in range(0, 10):
            random.Random(i).shuffle(steps)
            out_steps = sort_steps(steps, self.std_dep_map)
            enumerated = enumerate(step.sf_object for step in out_steps)
            positions = {name: pos for (pos, name) in enumerated}
            assert positions["Account"] < positions["Contact"], positions
            assert positions["Account"] < positions["Opportunity"], positions
            assert positions["Contact"] < positions["Opportunity"], positions
            assert positions["Opportunity"] < positions["Custom__c"], positions
            assert positions["Opportunity"] < positions["Custom__c"], positions
            assert positions["Custom__c"] < positions["Custom_2__c"], positions
            assert positions["Custom__c"] < positions["Custom_3__c"], positions

    def test_sort_steps__with_priorities(self):
        steps = [
            MappingStep(sf_object="Opportunity"),
            MappingStep(sf_object="Custom__c"),
            MappingStep(sf_object="Custom_2__c"),
            MappingStep(sf_object="Custom_3__c"),
            MappingStep(sf_object="Contact"),
            MappingStep(sf_object="Account"),
        ]

        for i in range(0, 10):
            random.Random(i).shuffle(steps)
            # add a high priority requirement that forces Custom_3__c to go before Account
            deps = DependencyMap(
                self.std_table_names,
                self.std_intertable_dependencies
                + [SObjDependency("Account", "Custom_3__c", "Custom_3__c", True)],
            )
            out_steps = sort_steps(steps, deps)
            enumerated = enumerate(step.sf_object for step in out_steps)
            positions = {name: pos for (pos, name) in enumerated}
            assert positions["Custom_3__c"] < positions["Account"], positions
            assert positions["Opportunity"] < positions["Custom__c"], positions
            assert positions["Opportunity"] < positions["Custom__c"], positions
            assert positions["Custom__c"] < positions["Custom_2__c"], positions
            assert positions["Custom__c"] < positions["Custom_3__c"], positions

    def test_sort_steps__with_priorities__2(self):
        steps = [
            MappingStep(sf_object="Opportunity"),
            MappingStep(sf_object="Custom__c"),
            MappingStep(sf_object="Custom_2__c"),
            MappingStep(sf_object="Custom_3__c"),
            MappingStep(sf_object="Contact"),
            MappingStep(sf_object="Account"),
        ]

        for i in range(0, 10):
            random.Random(i).shuffle(steps)
            # add a high priority requirement that forces Custom_2__c to go before Custom_3__c
            deps = DependencyMap(
                self.std_table_names,
                self.std_intertable_dependencies
                + [SObjDependency("Custom_3__c", "Custom_2__c", "Custom_2__c", True)],
            )
            out_steps = sort_steps(steps, deps)
            enumerated = enumerate(step.sf_object for step in out_steps)
            positions = {name: pos for (pos, name) in enumerated}
            assert positions["Custom_2__c"] < positions["Custom_3__c"], positions


class TestMergeMatchingSteps(MappingTransformTesterBase):
    def test_merge_matching_steps(self):
        steps = [
            MappingStep(sf_object="Account", fields={"recordtypeid": "recordtypeid"}),
            MappingStep(sf_object="Account", fields={"Name": "Name"}),
            MappingStep(sf_object="Contact", fields={"recordtype": "recordtype"}),
            MappingStep(sf_object="Contact", fields={"description": "description"}),
        ]
        out_steps = merge_matching_steps(steps, self.std_dep_map)
        by_name = {step.sf_object: step for step in out_steps}
        assert tuple(step.sf_object for step in out_steps) == ("Account", "Contact")
        assert tuple(by_name["Account"].fields) == ("recordtypeid", "Name")
        assert tuple(by_name["Contact"].fields) == ("recordtype", "description")


class TestRenameRecordTypeFields(MappingTransformTesterBase):
    def test_rename_record_type_fields(self):
        steps = [
            MappingStep(
                sf_object="Account", fields={"record_type_id": "record_type_id"}
            ),
            MappingStep(sf_object="Custom__c", fields={"recordtypeid": "recordtypeid"}),
            MappingStep(sf_object="Custom_2__c", fields={"recordtype": "recordtype"}),
            MappingStep(sf_object="Custom_3__c"),
        ]
        out_steps = rename_record_type_fields(steps, self.std_dep_map)

        by_name = {step.sf_object: step for step in out_steps}
        assert tuple(by_name.keys()) == (
            "Account",
            "Custom__c",
            "Custom_2__c",
            "Custom_3__c",
        )

        assert "RecordTypeId" in by_name["Account"].fields
        assert "RecordTypeId" in by_name["Custom__c"].fields
        assert "RecordTypeId" in by_name["Custom_2__c"].fields
        assert "RecordTypeId" not in by_name["Custom_3__c"].fields


class TestRecategorizeLookups(MappingTransformTesterBase):
    def test_recategorize_lookups(self):
        steps = [
            MappingStep(
                sf_object="Account",
                fields={
                    "ParentId": "ParentId",
                    "Name": "Name",
                    "Custom__c": "Custom__c",
                },
            ),
            MappingStep(
                sf_object="Contact",
                fields={"AccountId": "AccountId", "Firstname": "Firstname"},
            ),
            MappingStep(
                sf_object="Custom_2__c",
                fields={"Custom__c": "Custom__c", "Name": "Name"},
            ),
            MappingStep(
                sf_object="Custom_3__c",
                fields={"Name": "Name"},
            ),
        ]
        out_steps = recategorize_lookups(steps, self.std_dep_map)

        by_name = {step.sf_object: step for step in out_steps}
        assert tuple(by_name["Account"].fields) == ("Name",)
        assert tuple(by_name["Account"].lookups) == ("ParentId", "Custom__c")
        assert tuple(by_name["Contact"].fields) == ("Firstname",)
        assert tuple(by_name["Contact"].lookups) == ("AccountId",)
        assert tuple(by_name["Custom_2__c"].fields) == ("Name",)
        assert tuple(by_name["Custom_2__c"].lookups) == ("Custom__c",)
